@@ -1,12 +1,29 @@
 /*
  *  gensetup.c
  *
- *  $Id: gensetup.c,v 1.1 2001/06/07 00:50:20 source Exp $
+ *  $Id: gensetup.c,v 1.3 2003/08/22 14:28:40 source Exp $
  *
  *  The iODBC driver manager.
- * 
- *  Copyright (C) 2001 by OpenLink Software <iodbc@openlinksw.com>
+ *  
+ *  Copyright (C) 1999-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
+ *  This software is released under the terms of either of the following
+ *  licenses:
+ *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -20,6 +37,35 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include "gui.h"
@@ -58,20 +104,19 @@ addkeywords_to_list(GtkWidget* widget, LPCSTR attrs, TGENSETUP *gensetup_t)
 	  !strncasecmp (curr, "Description=", STRLEN ("Description=")))
 	continue;
 
-      data[0] = curr;
-
       if ((cour = strchr (curr, '=')))
-	{
-	  data[1] = cour + 1;
-	  *cour = 0;
-	  gtk_clist_append (GTK_CLIST (widget), data);
-	  *cour = '=';
-	}
+	     {
+		    *cour = '\0';
+          data[0] = curr;
+	       data[1] = cour + 1;
+	       gtk_clist_append (GTK_CLIST (widget), data);
+		    *cour = '=';
+	     }
       else
-	{
-	  data[1] = "";
-	  gtk_clist_append (GTK_CLIST (widget), data);
-	}
+	    {
+	      data[0] = "";
+	      gtk_clist_append (GTK_CLIST (widget), data);
+	    }
     }
 
   if (GTK_CLIST (widget)->rows > 0)
@@ -107,23 +152,30 @@ gensetup_add_clicked(GtkWidget* widget, TGENSETUP *gensetup_t)
   if (gensetup_t)
     {
       data[0] = gtk_entry_get_text (GTK_ENTRY (gensetup_t->key_entry));
-      if (!STRLEN (data[0]))
-	goto done;
-      data[1] = gtk_entry_get_text (GTK_ENTRY (gensetup_t->value_entry));
+      if (STRLEN (data[0]))
+        {
+		    data[1] = gtk_entry_get_text (GTK_ENTRY (gensetup_t->value_entry));
 
-      for (i = 0; i < GTK_CLIST (gensetup_t->key_list)->rows; i++)
-	{
-	  gtk_clist_get_text (GTK_CLIST (gensetup_t->key_list), i, 0, &szKey);
-	  if (!strcmp (szKey, data[0]))
-	    {
-	      gtk_clist_remove (GTK_CLIST (gensetup_t->key_list), i);
-	      break;
-	    }
-	}
+          /* Try to see if the keyword already exists */
+			 for (i = 0 ; i < GTK_CLIST (gensetup_t->key_list)->rows ; i++)
+			   {
+              gtk_clist_get_text (GTK_CLIST (gensetup_t->key_list), i, 0, &szKey);
+	           if(STREQ(data[0], szKey))
+				    goto done;
+				}
 
-      gtk_clist_append (GTK_CLIST (gensetup_t->key_list), data);
-      if (GTK_CLIST (gensetup_t->key_list)->rows > 0)
-	gtk_clist_sort (GTK_CLIST (gensetup_t->key_list));
+          /* An update operation */
+			 if (i < GTK_CLIST (gensetup_t->key_list)->rows)
+			   {
+				  gtk_clist_set_text (GTK_CLIST (gensetup_t->key_list), i, 1, data[1]);
+				}
+			 else if (STRLEN(data[1]))
+			   {
+              gtk_clist_append (GTK_CLIST (gensetup_t->key_list), data);
+				}
+		  }
+
+	   gtk_clist_sort (GTK_CLIST (gensetup_t->key_list));
 
     done:
       gtk_entry_set_text (GTK_ENTRY (gensetup_t->key_entry), "");
@@ -135,22 +187,28 @@ gensetup_add_clicked(GtkWidget* widget, TGENSETUP *gensetup_t)
 static void
 gensetup_update_clicked(GtkWidget* widget, TGENSETUP *gensetup_t)
 {
+  char *szKey;
   char *data[2];
+  int i;
 
-  if (gensetup_t && GTK_CLIST (gensetup_t->key_list)->selection != NULL)
+  if (gensetup_t)
     {
       data[0] = gtk_entry_get_text (GTK_ENTRY (gensetup_t->key_entry));
-      data[1] = gtk_entry_get_text (GTK_ENTRY (gensetup_t->value_entry));
-      gtk_clist_remove (GTK_CLIST (gensetup_t->key_list),
-	  GPOINTER_TO_INT (GTK_CLIST (gensetup_t->key_list)->selection->
-	      data));
+		if (STRLEN (data[0]))
+		  {
+          data[1] = gtk_entry_get_text (GTK_ENTRY (gensetup_t->value_entry));
 
-      if (STRLEN (data[0]))
-	{
-	  gtk_clist_append (GTK_CLIST (gensetup_t->key_list), data);
-	  if (GTK_CLIST (gensetup_t->key_list)->rows > 0)
-	    gtk_clist_sort (GTK_CLIST (gensetup_t->key_list));
-	}
+          if (GTK_CLIST (gensetup_t->key_list)->selection != NULL)
+		      i = GPOINTER_TO_INT (GTK_CLIST (gensetup_t->key_list)->selection->data);
+		    else i = 0;
+
+          /* An update operation */
+          if (i < GTK_CLIST (gensetup_t->key_list)->rows)
+		      {
+				  gtk_clist_set_text (GTK_CLIST (gensetup_t->key_list), i, 0, data[0]);
+				  gtk_clist_set_text (GTK_CLIST (gensetup_t->key_list), i, 1, data[1]);
+		      }
+		  }
 
       gtk_entry_set_text (GTK_ENTRY (gensetup_t->key_entry), "");
       gtk_entry_set_text (GTK_ENTRY (gensetup_t->value_entry), "");
@@ -194,72 +252,54 @@ static void
 gensetup_ok_clicked(GtkWidget* widget, TGENSETUP *gensetup_t)
 {
   char *curr, *cour, *szKey, *szValue;
-  int i = 0, size;
+  int i = 0, size = 0;
 
   if (gensetup_t)
     {
-      gensetup_t->connstr = (char *) malloc (sizeof (char) * (size =
-	      ((STRLEN (gtk_entry_get_text (GTK_ENTRY (gensetup_t->
-				  dsn_entry))) ?
-		      STRLEN (gtk_entry_get_text (GTK_ENTRY (gensetup_t->
-				  dsn_entry))) + STRLEN ("DSN=") +
-		      1 : STRLEN ("DSN=") + 1) +
-		  (STRLEN (gtk_entry_get_text (GTK_ENTRY (gensetup_t->
-				  comment_entry))) ?
-		      STRLEN (gtk_entry_get_text (GTK_ENTRY (gensetup_t->
-				  comment_entry))) + STRLEN ("Description=") +
-		      1 : STRLEN ("Description=") + 1) + 1)));
+	   /* What is the size of the block to malloc */
+      size += STRLEN (gtk_entry_get_text (GTK_ENTRY (gensetup_t->dsn_entry))) + STRLEN ("DSN=") + 1;
+      size += STRLEN (gtk_entry_get_text (GTK_ENTRY (gensetup_t->comment_entry))) + STRLEN ("Description=") + 1;
+      /* Malloc it */
+		if ((gensetup_t->connstr = (char*) malloc (++size)))
+		  {
+		    for (curr = STRCONN, cour = gensetup_t->connstr ;
+			   i < STRCONN_NB_TOKENS ; i++, curr += (STRLEN(curr) + 1))
+			   switch (i)
+				  {
+				    case 0:
+					   sprintf (cour, curr, gtk_entry_get_text (GTK_ENTRY (gensetup_t->dsn_entry)));
+						cour += (STRLEN (cour) + 1);
+					   break;
+					 case 1:
+					   sprintf (cour, curr, gtk_entry_get_text (GTK_ENTRY (gensetup_t->comment_entry)));
+						cour += (STRLEN (cour) + 1);
+					   break;
+				  };
 
-      if (gensetup_t->connstr)
-	{
-	  for (curr = STRCONN, cour = gensetup_t->connstr;
-	      i < STRCONN_NB_TOKENS; i++, curr += (STRLEN (curr) + 1))
-	    switch (i)
-	      {
-	      case 0:
-		sprintf (cour, curr,
-		    gtk_entry_get_text (GTK_ENTRY (gensetup_t->dsn_entry)));
-		cour += (STRLEN (cour) + 1);
-		break;
-	      case 1:
-		sprintf (cour, curr,
-		    gtk_entry_get_text (GTK_ENTRY (gensetup_t->
-			    comment_entry)));
-		cour += (STRLEN (cour) + 1);
-		break;
-	      };
+	       for (i = 0; i < GTK_CLIST (gensetup_t->key_list)->rows; i++)
+	        {
+	          gtk_clist_get_text (GTK_CLIST (gensetup_t->key_list), i, 0, &szKey);
+	          gtk_clist_get_text (GTK_CLIST (gensetup_t->key_list), i, 1, &szValue);
 
-	  for (i = 0; i < GTK_CLIST (gensetup_t->key_list)->rows; i++)
-	    {
-	      gtk_clist_get_text (GTK_CLIST (gensetup_t->key_list), i, 0,
-		  &szKey);
-	      gtk_clist_get_text (GTK_CLIST (gensetup_t->key_list), i, 1,
-		  &szValue);
-
-	      cour = (char *) gensetup_t->connstr;
-	      gensetup_t->connstr =
-		  (LPSTR) malloc (size + STRLEN (szKey) + STRLEN (szValue) +
-		  2);
-	      if (gensetup_t->connstr)
-		{
-		  memcpy (gensetup_t->connstr, cour, size);
-		  sprintf (gensetup_t->connstr + size - 1, "%s=%s", szKey,
-		      szValue);
-		  free (cour);
-		  size += STRLEN (szKey) + STRLEN (szValue) + 2;
-		}
-	      else
-		gensetup_t->connstr = cour;
-	    }
-
-	  gensetup_t->connstr[size - 1] = 0;
-	}
+             cour = gensetup_t->connstr;
+				 gensetup_t->connstr = (char*)malloc (size + STRLEN (szKey) + STRLEN(szValue) + 2);
+             if (gensetup_t->connstr)
+				   {
+					  memcpy (gensetup_t->connstr, cour, size);
+					  sprintf (gensetup_t->connstr + size - 1, "%s=%s", szKey, szValue);
+					  free (cour);
+					  size += STRLEN (szKey) + STRLEN (szValue) + 2;
+					}
+				 else
+				   gensetup_t->connstr = cour;
+			  }
+		  }
 
       gensetup_t->dsn_entry = gensetup_t->comment_entry = NULL;
       gensetup_t->key_list = NULL;
 
       gtk_signal_disconnect_by_func (GTK_OBJECT (gensetup_t->mainwnd),
-	  GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
+	    GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
       gtk_main_quit ();
       gtk_widget_destroy (gensetup_t->mainwnd);
     }
@@ -277,7 +317,7 @@ gensetup_cancel_clicked(GtkWidget* widget, TGENSETUP *gensetup_t)
       gensetup_t->key_list = NULL;
 
       gtk_signal_disconnect_by_func (GTK_OBJECT (gensetup_t->mainwnd),
-	  GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
+	     GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
       gtk_main_quit ();
       gtk_widget_destroy (gensetup_t->mainwnd);
     }
@@ -301,9 +341,10 @@ create_gensetup (HWND hwnd, LPCSTR dsn, LPCSTR attrs, BOOL add)
   GtkWidget *l_key, *l_value, *t_keyword, *t_value, *l_copyright;
   GtkWidget *vbuttonbox1, *b_add, *b_update, *l_keyword, *l_valeur;
   GtkWidget *dialog_action_area1, *hbuttonbox1, *b_ok, *b_cancel;
-  guint b_add_key, b_update_key, b_ok_key, b_cancel_key;
+  guint button_key;
   GtkAccelGroup *accel_group;
   TGENSETUP gensetup_t;
+  char buff[1024];
 
   if (hwnd == NULL || !GTK_IS_WIDGET (hwnd))
     return (LPSTR) attrs;
@@ -312,7 +353,8 @@ create_gensetup (HWND hwnd, LPCSTR dsn, LPCSTR attrs, BOOL add)
 
   gensetup = gtk_dialog_new ();
   gtk_object_set_data (GTK_OBJECT (gensetup), "gensetup", gensetup);
-  gtk_window_set_title (GTK_WINDOW (gensetup), "Generic ODBC Driver Setup");
+  sprintf (buff, "Setup of DSN %s ...", (dsn) ? dsn : "Unknown");
+  gtk_window_set_title (GTK_WINDOW (gensetup), buff);
   gtk_window_set_position (GTK_WINDOW (gensetup), GTK_WIN_POS_CENTER);
   gtk_window_set_modal (GTK_WINDOW (gensetup), TRUE);
   gtk_window_set_policy (GTK_WINDOW (gensetup), FALSE, FALSE, FALSE);
@@ -437,10 +479,10 @@ create_gensetup (HWND hwnd, LPCSTR dsn, LPCSTR attrs, BOOL add)
   gtk_widget_set_usize (vbuttonbox1, 85, 69);
 
   b_add = gtk_button_new_with_label ("");
-  b_add_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
+  button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
       szKeysButtons[0]);
   gtk_widget_add_accelerator (b_add, "clicked", accel_group,
-      b_add_key, GDK_MOD1_MASK, 0);
+      button_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_add);
   gtk_object_set_data_full (GTK_OBJECT (gensetup), "b_add", b_add,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -451,10 +493,10 @@ create_gensetup (HWND hwnd, LPCSTR dsn, LPCSTR attrs, BOOL add)
       'A', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_update = gtk_button_new_with_label ("");
-  b_update_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_update)->child),
+  button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_update)->child),
       szKeysButtons[1]);
   gtk_widget_add_accelerator (b_update, "clicked", accel_group,
-      b_update_key, GDK_MOD1_MASK, 0);
+      button_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_update);
   gtk_object_set_data_full (GTK_OBJECT (gensetup), "b_update", b_update,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -502,9 +544,9 @@ create_gensetup (HWND hwnd, LPCSTR dsn, LPCSTR attrs, BOOL add)
   gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbuttonbox1), 10);
 
   b_ok = gtk_button_new_with_label ("");
-  b_ok_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_ok)->child), "_Ok");
+  button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_ok)->child), "_Ok");
   gtk_widget_add_accelerator (b_ok, "clicked", accel_group,
-      b_ok_key, GDK_MOD1_MASK, 0);
+      button_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_ok);
   gtk_object_set_data_full (GTK_OBJECT (gensetup), "b_ok", b_ok,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -515,10 +557,10 @@ create_gensetup (HWND hwnd, LPCSTR dsn, LPCSTR attrs, BOOL add)
       'O', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_cancel = gtk_button_new_with_label ("");
-  b_cancel_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_cancel)->child),
+  button_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_cancel)->child),
       "_Cancel");
   gtk_widget_add_accelerator (b_cancel, "clicked", accel_group,
-      b_cancel_key, GDK_MOD1_MASK, 0);
+      button_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_cancel);
   gtk_object_set_data_full (GTK_OBJECT (gensetup), "b_cancel", b_cancel,
       (GtkDestroyNotify) gtk_widget_unref);
