@@ -1,14 +1,32 @@
 /*
  *  execute.c
  *
- *  $Id: execute.c,v 1.11 2001/06/04 14:01:25 source Exp $
+ *  $Id: execute.c,v 1.16 2003/09/03 08:44:47 source Exp $
  *
  *  Invoke a query
  *
  *  The iODBC driver manager.
  *  
  *  Copyright (C) 1995 by Ke Jin <kejin@empress.com> 
+ *  Copyright (C) 1996-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
+ *  This software is released under the terms of either of the following
+ *  licenses:
+ *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -22,12 +40,44 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <iodbc.h>
 
 #include <sql.h>
 #include <sqlext.h>
+#include <sqlucode.h>
+
+#include <unicode.h>
 
 #include <dlproc.h>
 
@@ -38,8 +88,8 @@
 
 #include <itrace.h>
 
-static void
-do_cursoropen (STMT_t FAR * pstmt)
+void
+_iodbcdm_do_cursoropen (STMT_t FAR * pstmt)
 {
   SQLRETURN retcode;
   SWORD ncol;
@@ -64,8 +114,8 @@ do_cursoropen (STMT_t FAR * pstmt)
 }
 
 
-SQLRETURN SQL_API
-SQLExecute (SQLHSTMT hstmt)
+static SQLRETURN
+SQLExecute_Internal (SQLHSTMT hstmt)
 {
   STMT (pstmt, hstmt);
   HPROC hproc = SQL_NULL_HPROC;
@@ -73,8 +123,6 @@ SQLExecute (SQLHSTMT hstmt)
 
   int sqlstat = en_00000;
 
-  ENTER_STMT (pstmt);
-
   /* check state */
   if (pstmt->asyn_on == en_NullProc)
     {
@@ -84,6 +132,7 @@ SQLExecute (SQLHSTMT hstmt)
 	  sqlstat = en_S1010;
 	  break;
 
+	case en_stmt_executed_with_info:
 	case en_stmt_executed:
 	  if (!pstmt->prep_state)
 	    {
@@ -139,7 +188,7 @@ SQLExecute (SQLHSTMT hstmt)
     {
       PUSHSQLERR (pstmt->herr, sqlstat);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_Execute,
@@ -159,7 +208,7 @@ SQLExecute (SQLHSTMT hstmt)
 
 	case SQL_STILL_EXECUTING:
 	default:
-	  LEAVE_STMT (pstmt, retcode);
+	  return retcode;
 	}
     }
 
@@ -169,8 +218,11 @@ SQLExecute (SQLHSTMT hstmt)
       switch (retcode)
 	{
 	case SQL_SUCCESS:
+	  _iodbcdm_do_cursoropen (pstmt);
+	  break;
+
 	case SQL_SUCCESS_WITH_INFO:
-	  do_cursoropen (pstmt);
+	  pstmt->state = en_stmt_executed_with_info;
 	  break;
 
 	case SQL_NEED_DATA:
@@ -214,20 +266,36 @@ SQLExecute (SQLHSTMT hstmt)
       break;
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLExecDirect (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
+SQLExecute (SQLHSTMT hstmt)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLExecute (TRACE_ENTER, hstmt));
+
+  retcode = SQLExecute_Internal (hstmt);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLExecute (TRACE_LEAVE, hstmt));
+}
+
+
+SQLRETURN SQL_API
+SQLExecDirect_Internal (SQLHSTMT hstmt,
+    SQLPOINTER szSqlStr,
+    SQLINTEGER cbSqlStr,
+    SQLCHAR waMode)
 {
   STMT (pstmt, hstmt);
+  CONN (pdbc, pstmt->hdbc);
+  ENV_t FAR *penv = pdbc->henv;
   HPROC hproc = SQL_NULL_HPROC;
-
-  int sqlstat = en_00000;
   SQLRETURN retcode = SQL_SUCCESS;
-
-  ENTER_STMT (pstmt);
+  int sqlstat = en_00000;
+  void * _SqlStr = NULL;
 
   /* check arguments */
   if (szSqlStr == NULL)
@@ -243,7 +311,7 @@ SQLExecDirect (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
     {
       PUSHSQLERR (pstmt->herr, sqlstat);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* check state */
@@ -275,20 +343,41 @@ SQLExecDirect (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
     {
       PUSHSQLERR (pstmt->herr, sqlstat);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
-  hproc = _iodbcdm_getproc (pstmt->hdbc, en_ExecDirect);
+  if ((penv->unicode_driver && waMode != 'W')
+      || (!penv->unicode_driver && waMode == 'W'))
+    {
+      if (waMode != 'W')
+        {
+        /* ansi=>unicode*/
+          _SqlStr = _iodbcdm_conv_param_A2W(pstmt, 0, szSqlStr, cbSqlStr);
+        }
+      else
+        {
+        /* unicode=>ansi*/
+          _SqlStr = _iodbcdm_conv_param_W2A(pstmt, 0, szSqlStr, cbSqlStr);
+        }
+      szSqlStr = _SqlStr;
+      cbSqlStr = SQL_NTS;
+    }
+
+  CALL_UDRIVER(pstmt->hdbc, pstmt, retcode, hproc, penv->unicode_driver,
+    en_ExecDirect, (
+       pstmt->dhstmt,
+       szSqlStr,
+       cbSqlStr));
 
   if (hproc == SQL_NULL_HPROC)
     {
+      _iodbcdm_FreeStmtParams(pstmt);
       PUSHSQLERR (pstmt->herr, en_IM001);
-
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
-  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_ExecDirect,
-      (pstmt->dhstmt, szSqlStr, cbSqlStr));
+  if (retcode != SQL_STILL_EXECUTING)
+    _iodbcdm_FreeStmtParams(pstmt);
 
   /* stmt state transition */
   if (pstmt->asyn_on == en_ExecDirect)
@@ -304,7 +393,7 @@ SQLExecDirect (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
 
 	case SQL_STILL_EXECUTING:
 	default:
-	  LEAVE_STMT (pstmt, retcode);
+	  return retcode;
 	}
     }
 
@@ -313,8 +402,13 @@ SQLExecDirect (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
       switch (retcode)
 	{
 	case SQL_SUCCESS:
+	  _iodbcdm_do_cursoropen (pstmt);
+	  pstmt->prep_state = 1;
+	  break;
+
 	case SQL_SUCCESS_WITH_INFO:
-	  do_cursoropen (pstmt);
+	  pstmt->state = en_stmt_executed_with_info;
+	  pstmt->prep_state = 1;
 	  break;
 
 	case SQL_NEED_DATA:
@@ -337,26 +431,63 @@ SQLExecDirect (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
 	}
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLPutData (SQLHSTMT hstmt, SQLPOINTER rgbValue, SQLINTEGER cbValue)
+SQLExecDirect (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLExecDirect (TRACE_ENTER, hstmt, szSqlStr, cbSqlStr));
+
+  retcode = SQLExecDirect_Internal(hstmt, szSqlStr, cbSqlStr, 'A');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLExecDirect (TRACE_LEAVE, hstmt, szSqlStr, cbSqlStr));
+}
+
+
+SQLRETURN SQL_API
+SQLExecDirectA (SQLHSTMT hstmt, SQLCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLExecDirect (TRACE_ENTER, hstmt, szSqlStr, cbSqlStr));
+
+  retcode = SQLExecDirect_Internal(hstmt, szSqlStr, cbSqlStr, 'A');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLExecDirect (TRACE_LEAVE, hstmt, szSqlStr, cbSqlStr));
+}
+
+
+SQLRETURN SQL_API
+SQLExecDirectW (SQLHSTMT hstmt, SQLWCHAR FAR * szSqlStr, SQLINTEGER cbSqlStr)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLExecDirectW (TRACE_ENTER, hstmt, szSqlStr, cbSqlStr));
+
+  retcode = SQLExecDirect_Internal(hstmt, szSqlStr, cbSqlStr, 'W');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLExecDirectW (TRACE_LEAVE, hstmt, szSqlStr, cbSqlStr));
+}
+
+
+static SQLRETURN
+SQLPutData_Internal (SQLHSTMT hstmt, SQLPOINTER rgbValue, SQLINTEGER cbValue)
 {
   STMT (pstmt, hstmt);
   HPROC hproc;
   SQLRETURN retcode;
 
-  ENTER_STMT (pstmt);
-
   /* check argument value */
   if (rgbValue == NULL &&
       (cbValue != SQL_DEFAULT_PARAM && cbValue != SQL_NULL_DATA))
     {
       PUSHSQLERR (pstmt->herr, en_S1009);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* check state */
@@ -366,14 +497,14 @@ SQLPutData (SQLHSTMT hstmt, SQLPOINTER rgbValue, SQLINTEGER cbValue)
 	{
 	  PUSHSQLERR (pstmt->herr, en_S1010);
 
-	  LEAVE_STMT (pstmt, SQL_ERROR);
+	  return SQL_ERROR;
 	}
     }
   else if (pstmt->asyn_on != en_PutData)
     {
       PUSHSQLERR (pstmt->herr, en_S1010);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* call driver */
@@ -383,7 +514,7 @@ SQLPutData (SQLHSTMT hstmt, SQLPOINTER rgbValue, SQLINTEGER cbValue)
     {
       PUSHSQLERR (pstmt->herr, en_IM001);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_PutData,
@@ -402,7 +533,7 @@ SQLPutData (SQLHSTMT hstmt, SQLPOINTER rgbValue, SQLINTEGER cbValue)
 
 	case SQL_STILL_EXECUTING:
 	default:
-	  LEAVE_STMT (pstmt, retcode);
+	  return retcode;
 	}
     }
 
@@ -448,19 +579,30 @@ SQLPutData (SQLHSTMT hstmt, SQLPOINTER rgbValue, SQLINTEGER cbValue)
       break;
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLParamData (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
+SQLPutData (SQLHSTMT hstmt, SQLPOINTER rgbValue, SQLINTEGER cbValue)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLPutData (TRACE_ENTER, hstmt, rgbValue, cbValue));
+
+  retcode = SQLPutData_Internal (hstmt, rgbValue, cbValue);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLPutData (TRACE_LEAVE, hstmt, rgbValue, cbValue));
+}
+
+
+static SQLRETURN
+SQLParamData_Internal (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
 {
   STMT (pstmt, hstmt);
   HPROC hproc;
   SQLRETURN retcode;
 
-  ENTER_STMT (pstmt);
-
   /* check argument */
 
   /* check state */
@@ -470,14 +612,14 @@ SQLParamData (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
 	{
 	  PUSHSQLERR (pstmt->herr, en_S1010);
 
-	  LEAVE_STMT (pstmt, SQL_ERROR);
+	  return SQL_ERROR;
 	}
     }
   else if (pstmt->asyn_on != en_ParamData)
     {
       PUSHSQLERR (pstmt->herr, en_S1010);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* call driver */
@@ -487,7 +629,7 @@ SQLParamData (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
     {
       PUSHSQLERR (pstmt->herr, en_IM001);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_ParamData,
@@ -506,13 +648,13 @@ SQLParamData (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
 
 	case SQL_STILL_EXECUTING:
 	default:
-	  LEAVE_STMT (pstmt, retcode);
+	  return retcode;
 	}
     }
 
   if (pstmt->state < en_stmt_needdata)
     {
-      LEAVE_STMT (pstmt, retcode);
+      return retcode;
     }
 
   switch (retcode)
@@ -557,7 +699,7 @@ SQLParamData (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
 
 	    case en_ExecDirect:
 	    case en_Execute:
-	      do_cursoropen (pstmt);
+	      _iodbcdm_do_cursoropen (pstmt);
 	      break;
 
 	    default:
@@ -579,23 +721,34 @@ SQLParamData (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
       break;
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLNumParams (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
+SQLParamData (SQLHSTMT hstmt, SQLPOINTER FAR * prgbValue)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLParamData (TRACE_ENTER, hstmt, prgbValue));
+
+  retcode = SQLParamData_Internal (hstmt, prgbValue);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLParamData (TRACE_LEAVE, hstmt, prgbValue));
+}
+
+
+static SQLRETURN
+SQLNumParams_Internal (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
 {
   STMT (pstmt, hstmt);
   HPROC hproc;
   SQLRETURN retcode;
 
-  ENTER_STMT (pstmt);
-
   /* check argument */
   if (!pcpar)
     {
-      LEAVE_STMT (pstmt, SQL_SUCCESS);
+      return SQL_SUCCESS;
     }
 
   /* check state */
@@ -608,7 +761,7 @@ SQLNumParams (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
 	case en_stmt_mustput:
 	case en_stmt_canput:
 	  PUSHSQLERR (pstmt->herr, en_S1010);
-	  LEAVE_STMT (pstmt, SQL_ERROR);
+	  return SQL_ERROR;
 
 	default:
 	  break;
@@ -618,7 +771,7 @@ SQLNumParams (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
     {
       PUSHSQLERR (pstmt->herr, en_S1010);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* call driver */
@@ -628,7 +781,7 @@ SQLNumParams (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
     {
       PUSHSQLERR (pstmt->herr, en_IM001);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_NumParams,
@@ -645,7 +798,7 @@ SQLNumParams (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
 	  break;
 
 	default:
-	  LEAVE_STMT (pstmt, retcode);
+	  return retcode;
 	}
     }
 
@@ -654,29 +807,43 @@ SQLNumParams (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
       pstmt->asyn_on = en_NumParams;
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLDescribeParam (SQLHSTMT hstmt,
+SQLNumParams (SQLHSTMT hstmt, SQLSMALLINT FAR * pcpar)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLNumParams (TRACE_ENTER, hstmt, pcpar));
+
+  retcode = SQLNumParams_Internal (hstmt, pcpar);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLNumParams (TRACE_LEAVE, hstmt, pcpar));
+}
+
+
+static SQLRETURN
+SQLDescribeParam_Internal (SQLHSTMT hstmt,
     SQLUSMALLINT ipar,
     SQLSMALLINT FAR * pfSqlType,
     SQLUINTEGER FAR * pcbColDef,
     SQLSMALLINT FAR * pibScale, SQLSMALLINT FAR * pfNullable)
 {
   STMT (pstmt, hstmt);
+  CONN (pdbc, pstmt->hdbc);
+  GENV_t FAR *genv = pdbc->genv;
+
   HPROC hproc;
   SQLRETURN retcode;
 
-  ENTER_STMT (pstmt);
-
   /* check argument */
   if (ipar == 0)
     {
       PUSHSQLERR (pstmt->herr, en_S1093);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* check state */
@@ -689,7 +856,7 @@ SQLDescribeParam (SQLHSTMT hstmt,
 	case en_stmt_mustput:
 	case en_stmt_canput:
 	  PUSHSQLERR (pstmt->herr, en_S1010);
-	  LEAVE_STMT (pstmt, SQL_ERROR);
+	  return SQL_ERROR;
 
 	default:
 	  break;
@@ -699,7 +866,7 @@ SQLDescribeParam (SQLHSTMT hstmt,
     {
       PUSHSQLERR (pstmt->herr, en_S1010);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* call driver */
@@ -709,12 +876,19 @@ SQLDescribeParam (SQLHSTMT hstmt,
     {
       PUSHSQLERR (pstmt->herr, en_IM001);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_DescribeParam,
       (pstmt->dhstmt, ipar, pfSqlType, pcbColDef, pibScale, pfNullable));
 
+  /*
+   *  Convert sql type to ODBC version of application
+   */
+  if (SQL_SUCCEEDED(retcode) && pfSqlType)
+    *pfSqlType = _iodbcdm_map_sql_type (*pfSqlType, genv->odbc_ver);
+
+
   /* state transition */
   if (pstmt->asyn_on == en_DescribeParam)
     {
@@ -726,7 +900,7 @@ SQLDescribeParam (SQLHSTMT hstmt,
 	  break;
 
 	default:
-	  LEAVE_STMT (pstmt, retcode);
+	  return retcode;
 	}
     }
 
@@ -735,5 +909,29 @@ SQLDescribeParam (SQLHSTMT hstmt,
       pstmt->asyn_on = en_DescribeParam;
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
+}
+
+
+SQLRETURN SQL_API
+SQLDescribeParam (
+  SQLHSTMT		  hstmt,
+  SQLUSMALLINT		  ipar,
+  SQLSMALLINT FAR	* pfSqlType,
+  SQLUINTEGER FAR	* pcbColDef,
+  SQLSMALLINT FAR	* pibScale,
+  SQLSMALLINT FAR	* pfNullable)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLDescribeParam (TRACE_ENTER,
+      hstmt, ipar, pfSqlType,
+      pcbColDef, pibScale, pfNullable));
+
+  retcode = SQLDescribeParam_Internal ( hstmt, ipar, pfSqlType,
+      pcbColDef, pibScale, pfNullable);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLDescribeParam (TRACE_LEAVE,
+      hstmt, ipar, pfSqlType,
+      pcbColDef, pibScale, pfNullable));
 }
