@@ -1,14 +1,32 @@
 /*
  *  hdbc.c
  *
- *  $Id: hdbc.c,v 1.14 2001/06/04 14:01:25 source Exp $
+ *  $Id: hdbc.c,v 1.22 2003/09/03 08:44:47 source Exp $
  *
  *  Data source connect object management functions
  *
  *  The iODBC driver manager.
  *  
  *  Copyright (C) 1995 by Ke Jin <kejin@empress.com> 
+ *  Copyright (C) 1996-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
+ *  This software is released under the terms of either of the following
+ *  licenses:
+ *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -22,12 +40,45 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <iodbc.h>
 
 #include <sql.h>
 #include <sqlext.h>
+#include <sqlucode.h>
+#include <iodbcext.h>
+
+#include <unicode.h>
 
 #include <dlproc.h>
 
@@ -39,26 +90,78 @@
 #include <itrace.h>
 #include <stdio.h>
 
-SQLRETURN SQL_API
-SQLAllocConnect (
+
+extern SQLRETURN _iodbcdm_driverunload ();
+
+static SQLRETURN
+_iodbcdm_drvopt_store (SQLHDBC hdbc, SQLUSMALLINT fOption, SQLUINTEGER vParam,
+	SQLCHAR waMode)
+{
+  CONN (pdbc, hdbc);
+  DRVOPT *popt;
+
+  /*
+   *  Check if this option is already registered
+   */
+  for (popt = pdbc->drvopt; popt != NULL; popt = popt->next)
+    {
+      if (popt->Option == fOption)
+	break;
+    }
+
+  /*
+   *  New option
+   */
+  if (popt == NULL)
+    {
+      if ((popt = (DRVOPT *) MEM_ALLOC (sizeof (DRVOPT))) == NULL)
+	return SQL_ERROR;
+
+      popt->Option = fOption;
+      popt->next = pdbc->drvopt;
+      pdbc->drvopt = popt;
+    }
+
+  /*
+   *  Store the value
+   */
+  popt->Param = vParam;
+  popt->waMode = waMode;
+
+  return SQL_SUCCESS;
+}
+
+
+static SQLRETURN
+_iodbcdm_drvopt_free (SQLHDBC hdbc)
+{
+  CONN (pdbc, hdbc);
+  DRVOPT *popt;
+
+  popt = pdbc->drvopt;
+  while (popt != NULL)
+    {
+      DRVOPT *tmp = popt->next;
+      free (popt);
+      popt = tmp;
+    }
+  pdbc->drvopt = NULL;
+
+  return SQL_SUCCESS;
+}
+
+
+SQLRETURN 
+SQLAllocConnect_Internal (
     SQLHENV henv,
     SQLHDBC FAR * phdbc)
 {
   GENV (genv, henv);
   DBC_t FAR *pdbc;
 
-  ODBC_LOCK ();
-  if (!IS_VALID_HENV (genv))
-    {
-      ODBC_UNLOCK ();
-      return SQL_INVALID_HANDLE;
-    }
-  CLEAR_ERRORS (genv);
-
   if (phdbc == NULL)
     {
       PUSHSQLERR (genv->herr, en_S1009);
-      ODBC_UNLOCK ();
       return SQL_ERROR;
     }
 
@@ -69,7 +172,6 @@ SQLAllocConnect (
       *phdbc = SQL_NULL_HDBC;
 
       PUSHSQLERR (genv->herr, en_S1001);
-      ODBC_UNLOCK ();
       return SQL_ERROR;
     }
   pdbc->rc = 0;
@@ -94,10 +196,9 @@ SQLAllocConnect (
   pdbc->herr = SQL_NULL_HERR;
   pdbc->dhdbc = SQL_NULL_HDBC;
   pdbc->state = en_dbc_allocated;
-  pdbc->trace = 0;
-  pdbc->tstm = NULL;
-  pdbc->tfile = NULL;
+  pdbc->drvopt = NULL;
   pdbc->dbc_cip = 0;
+  pdbc->err_rec = 0;
 
   /* set connect options to default values */
   pdbc->access_mode = SQL_MODE_DEFAULT;
@@ -113,31 +214,47 @@ SQLAllocConnect (
 
   *phdbc = (SQLHDBC) pdbc;
 
-  ODBC_UNLOCK ();
   return SQL_SUCCESS;
 }
 
 
 SQLRETURN SQL_API
-SQLFreeConnect (SQLHDBC hdbc)
+SQLAllocConnect (SQLHENV henv, SQLHDBC FAR * phdbc)
+{
+  GENV (genv, henv);
+  SQLRETURN retcode = SQL_SUCCESS;
+
+  ODBC_LOCK ();
+  if (!IS_VALID_HENV (genv))
+    {
+      ODBC_UNLOCK ();
+      return SQL_INVALID_HANDLE;
+    }
+  CLEAR_ERRORS (genv);
+
+  TRACE (trace_SQLAllocConnect (TRACE_ENTER, henv, phdbc));
+
+  retcode = SQLAllocConnect_Internal (henv, phdbc);
+
+  TRACE (trace_SQLAllocConnect (TRACE_LEAVE, henv, phdbc));
+
+  ODBC_UNLOCK ();
+
+  return SQL_SUCCESS;
+}
+
+
+SQLRETURN
+SQLFreeConnect_Internal (SQLHDBC hdbc)
 {
   GENV_t FAR *genv;
   CONN (pdbc, hdbc);
   DBC_t FAR *tpdbc;
 
-  ODBC_LOCK ();
-  if (!IS_VALID_HDBC (pdbc))
-    {
-      ODBC_UNLOCK ();
-      return SQL_INVALID_HANDLE;
-    }
-  CLEAR_ERRORS (pdbc);
-
   /* check state */
   if (pdbc->state != en_dbc_allocated)
     {
       PUSHSQLERR (pdbc->herr, en_S1010);
-      ODBC_UNLOCK ();
       return SQL_ERROR;
     }
 
@@ -161,32 +278,44 @@ SQLFreeConnect (SQLHDBC hdbc)
   /* free this dbc */
   _iodbcdm_driverunload (pdbc);
 
-  if (pdbc->tfile)
-    {
-      MEM_FREE (pdbc->tfile);
-    }
-
-  _iodbcdm_SetConnectOption (hdbc, SQL_OPT_TRACE, SQL_OPT_TRACE_OFF);
+  /* free driver connect options */
+  _iodbcdm_drvopt_free (pdbc);
+   
+  _iodbcdm_SetConnectOption (hdbc, SQL_OPT_TRACE, SQL_OPT_TRACE_OFF, 'A');
 
   /*
    *  Invalidate this handle
    */
   pdbc->type = 0;
 
-  MEM_FREE (pdbc);
-
-  ODBC_UNLOCK ();
   return SQL_SUCCESS;
 }
 
 
 SQLRETURN SQL_API
+SQLFreeConnect (SQLHDBC hdbc)
+{
+  ENTER_HDBC (hdbc, 1,
+    trace_SQLFreeConnect (TRACE_ENTER, hdbc));
+
+  retcode = SQLFreeConnect_Internal (hdbc);
+
+  LEAVE_HDBC (hdbc, 1,
+    trace_SQLFreeConnect (TRACE_LEAVE, hdbc);
+    MEM_FREE(hdbc);
+  );
+}
+
+
+SQLRETURN
 _iodbcdm_SetConnectOption (
     SQLHDBC hdbc,
     SQLUSMALLINT fOption,
-    SQLUINTEGER vParam)
+    SQLUINTEGER vParam,
+    SQLCHAR waMode)
 {
   CONN (pdbc, hdbc);
+  ENV_t FAR *penv = pdbc->henv;
   STMT_t FAR *pstmt;
   HPROC hproc = SQL_NULL_HPROC;
   int sqlstat = en_00000;
@@ -211,22 +340,20 @@ _iodbcdm_SetConnectOption (
 	{
 	  /* This two options are only meaningful
 	   * for specified driver. So, has to be
-	   * set after a dirver has been loaded.
+	   * set after a driver has been loaded.
 	   */
 	  sqlstat = en_08003;
 	  break;
 	}
 
+      /*
+       *  An option only meaningful for the driver is passed before the 
+       *  driver was actually loaded. We save it here and pass it onto 
+       *  the driver at a later stage.
+       */
       if (fOption >= SQL_CONNECT_OPT_DRVR_START && pdbc->henv == SQL_NULL_HENV)
-	/* An option only meaningful for drivers
-	 * is passed before loading a driver.
-	 * We classify this as an invalid option error.
-	 * This is not documented by MS SDK guide.
-	 */
-	{
-	  sqlstat = en_S1092;
-	  break;
-	}
+        _iodbcdm_drvopt_store (hdbc, fOption, vParam, waMode);
+
       break;
 
     case en_dbc_needdata:
@@ -274,95 +401,20 @@ _iodbcdm_SetConnectOption (
      * environment).
      */
     {
-      switch (vParam)
+      switch ((SQLUINTEGER)vParam)
 	{
 	case SQL_OPT_TRACE_ON:
-	  if (pdbc->tfile == NULL)
-	    {
-	      pdbc->tfile = (char FAR *) MEM_ALLOC (1 +
-		  STRLEN (SQL_OPT_TRACE_FILE_DEFAULT));
-
-	      if (pdbc->tfile == NULL)
-		{
-		  PUSHSQLERR (pdbc->herr, en_S1001);
-
-		  return SQL_ERROR;
-		}
-
-	      STRCPY (pdbc->tfile, SQL_OPT_TRACE_FILE_DEFAULT);
-	    }
-
-	  if (pdbc->tstm == NULL)
-	    {
-
-#if	defined(stderr) && defined(stdout)
-	      if (STREQ (pdbc->tfile, "stderr"))
-		{
-		  pdbc->tstm = stderr;
-		}
-	      else if (STREQ (pdbc->tfile, "stdout"))
-		{
-		  pdbc->tstm = stdout;
-		}
-	      else
-#endif
-
-		{
-#if defined (UNIX)
-		  /* 
-		   *  As this is a security risk, we refuse root to start
-		   *  a trace log
-		   */
-		  if (geteuid() == 0)
-		    {
-		      pdbc->tstm = NULL;
-		      pdbc->trace = 0;
-		      return SQL_SUCCESS;
-		    }
-		  else
-		    {
-		      pdbc->tstm = fopen (pdbc->tfile, "w");
-		    }
-#else
-		  pdbc->tstm = fopen (pdbc->tfile, "w");
-#endif
-		}
-
-	      if (pdbc->tstm)
-		{
-		  pdbc->trace = 1;
-		}
-	      else
-		{
-		  pdbc->trace = 0;
-
-		  sqlstat = en_IM013;
-		  retcode = SQL_ERROR;
-		}
-	    }
+	  trace_start ();
 	  break;
 
 	case SQL_OPT_TRACE_OFF:
-	  if (pdbc->trace && pdbc->tstm)
-	    {
-
-#if	defined(stderr) && defined(stdout)
-	      if (stderr != (FILE FAR *) (pdbc->tstm)
-		  && stdout != (FILE FAR *) (pdbc->tstm))
-#endif
-
-		{
-		  fclose (pdbc->tstm);
-		}
-	    }
-	  pdbc->tstm = NULL;
-	  pdbc->trace = 0;
+	  trace_stop ();
 	  break;
 
 	default:
-	  PUSHSQLERR (pdbc->herr, en_S1009);
-	  retcode = SQL_ERROR;
-	}
+	  PUSHSQLERR (pdbc->herr, en_HY024);
+	  return SQL_ERROR;
+	}                              	
 
       if (sqlstat != en_00000)
 	{
@@ -381,41 +433,35 @@ _iodbcdm_SetConnectOption (
      * and only meaningful for driver manager. 
      */
     {
-      if (vParam == 0UL || ((char FAR *) vParam)[0] == 0)
+      SQLCHAR *_vParam;
+      SQLCHAR *tmp = NULL;
+
+      if (((char FAR *)vParam) == NULL 
+          || (waMode != 'W' && ((char FAR *) vParam)[0] == '\0')
+          || (waMode == 'W' && ((wchar_t FAR *) vParam)[0] == L'\0'))
 	{
 	  PUSHSQLERR (pdbc->herr, en_S1009);
-
 	  return SQL_ERROR;
 	}
 
-      if (pdbc->tfile && STREQ (pdbc->tfile, vParam))
-	{
-	  return SQL_SUCCESS;
-	}
+      _vParam = (SQLCHAR *)vParam;
+      if (waMode == 'W')
+        {
+          if ((_vParam = tmp = dm_SQL_WtoU8((wchar_t *)vParam, SQL_NTS)) == NULL)
+	    {
+              PUSHSQLERR (pdbc->herr, en_S1001);
+              return SQL_ERROR;
+            }
+        }
 
-      if (pdbc->trace)
+      if (ODBCSharedTraceFlag)
 	{
-	  PUSHSQLERR (pdbc->herr, en_IM014);
-
+	  PUSHSQLERR (pdbc->herr, en_IM013);
 	  return SQL_ERROR;
 	}
 
-      if (pdbc->tfile)
-	{
-	  MEM_FREE (pdbc->tfile);
-	}
-
-      pdbc->tfile = (char FAR *) MEM_ALLOC (1 + STRLEN (vParam));
-
-      if (pdbc->tfile == NULL)
-	{
-	  PUSHSQLERR (pdbc->herr, en_S1001);
-
-	  return SQL_ERROR;
-	}
-
-      STRCPY (pdbc->tfile, vParam);
-
+      trace_set_filename ((char *) _vParam);
+      MEM_FREE (tmp);
       return SQL_SUCCESS;
     }
 
@@ -426,9 +472,57 @@ _iodbcdm_SetConnectOption (
        * and delay the setting process until the connection 
        * been established.  
        */
+     void * _vParam = NULL;
+     int procid = 0;
+
+     if ((penv->unicode_driver && waMode != 'W') 
+        || (!penv->unicode_driver && waMode == 'W'))
+      {
+        switch (fOption)
+          {
+          case SQL_ATTR_TRACEFILE:
+          case SQL_CURRENT_QUALIFIER:
+          case SQL_TRANSLATE_DLL:
+          case SQL_APPLICATION_NAME:
+          case SQL_COPT_SS_ENLIST_IN_DTC:
+          case SQL_COPT_SS_PERF_QUERY_LOG:
+          case SQL_COPT_SS_PERF_DATA_LOG:
+          case SQL_CURRENT_SCHEMA:
+            if (waMode != 'W')
+              {
+              /* ansi=>unicode*/
+                _vParam = dm_SQL_A2W((SQLCHAR *)vParam, SQL_NTS);
+              }
+            else
+              {
+              /* unicode=>ansi*/
+                _vParam = dm_SQL_W2A((SQLWCHAR *)vParam, SQL_NTS);
+              }
+            vParam = (SQLUINTEGER)_vParam;
+            break;
+          }
+      }
 
 #if (ODBCVER >= 0x0300)
-      hproc = _iodbcdm_getproc (pdbc, en_SetConnectAttr);
+     if (penv->unicode_driver)
+       {
+         /* SQL_XXX_W */
+         if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectAttrW))
+             != SQL_NULL_HPROC)
+           procid = en_SetConnectAttrW;
+       }
+     else
+       {
+         /* SQL_XXX */
+         /* SQL_XXX_A */
+         if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectAttr))
+             != SQL_NULL_HPROC)
+           procid = en_SetConnectAttr;
+         else
+         if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectAttrA))
+             != SQL_NULL_HPROC)
+           procid = en_SetConnectAttrA;
+       }
 
       if (hproc != SQL_NULL_HPROC)
 	{
@@ -443,7 +537,7 @@ _iodbcdm_SetConnectOption (
 	    case SQL_ATTR_QUIET_MODE:
 	    case SQL_ATTR_TRANSLATE_OPTION:
 	    case SQL_ATTR_TXN_ISOLATION:
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_SetConnectAttr,
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
 		  (pdbc->dhdbc, fOption, vParam, 0));
 	      break;
 
@@ -454,30 +548,50 @@ _iodbcdm_SetConnectOption (
 	    case SQL_ATTR_CONNECTION_TIMEOUT:
 	    case SQL_ATTR_METADATA_ID:
 	      PUSHSQLERR (pdbc->herr, en_IM001);
+	      MEM_FREE(_vParam);
 	      return SQL_ERROR;
 
 	    default:		/* string & driver defined */
 
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_SetConnectAttr,
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
 		  (pdbc->dhdbc, fOption, vParam, SQL_NTS));
 	    }
 	}
       else
 #endif
 	{
-	  hproc = _iodbcdm_getproc (pdbc, en_SetConnectOption);
+          if (penv->unicode_driver)
+            {
+             /* SQL_XXX_W */
+             if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectOptionW))
+                 != SQL_NULL_HPROC)
+               procid = en_SetConnectOptionW;
+            }
+          else
+            {
+             /* SQL_XXX */
+             /* SQL_XXX_A */
+             if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectOption))
+                != SQL_NULL_HPROC)
+               procid = en_SetConnectOption;
+             else
+             if ((hproc = _iodbcdm_getproc (pdbc, en_SetConnectOptionA))
+                != SQL_NULL_HPROC)
+               procid = en_SetConnectOptionA;
+            }
 
 	  if (hproc == SQL_NULL_HPROC)
 	    {
 	      PUSHSQLERR (pdbc->herr, en_IM001);
-
+	      MEM_FREE(_vParam);
 	      return SQL_ERROR;
 	    }
 
-	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_SetConnectOption,
+	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
 	      (pdbc->dhdbc, fOption, vParam));
 	}
 
+      MEM_FREE(_vParam);
       if (retcode != SQL_SUCCESS && retcode != SQL_SUCCESS_WITH_INFO)
 	{
 	  return retcode;
@@ -485,12 +599,12 @@ _iodbcdm_SetConnectOption (
     }
 
   /* 
-   * Now, either driver's odbc call was successed or
+   * Now, either driver's odbc call was successful or
    * driver has not been loaded yet. In the first case, we
    * need flip flag for(such as access_mode, autocommit, ...)
    * for our finit state machine. While in the second case, 
    * we need save option values(such as current_qualifier, ...)
-   * for delaied setting. So, ...
+   * for delayed setting. So, ...
    */
 
   /* No matter what state we are(i.e. allocated or connected, ..)
@@ -514,7 +628,7 @@ _iodbcdm_SetConnectOption (
     }
 
   /* Only 'allocated' state is possible here, and we need to
-   * save the options for delaied setting.
+   * save the options for delayed setting.
    */
   switch (fOption)
     {
@@ -527,13 +641,14 @@ _iodbcdm_SetConnectOption (
       if (vParam == 0UL)
 	{
 	  pdbc->current_qualifier = NULL;
-
 	  break;
 	}
 
-      pdbc->current_qualifier
-	  = (char FAR *) MEM_ALLOC (
-	  STRLEN (vParam) + 1);
+      if (waMode != 'W')
+        pdbc->current_qualifier = (wchar_t *) MEM_ALLOC (STRLEN (vParam) + 1);
+      else
+        pdbc->current_qualifier = (wchar_t *) MEM_ALLOC((WCSLEN (vParam) + 1) 
+        	* sizeof(wchar_t));
 
       if (pdbc->current_qualifier == NULL)
 	{
@@ -541,7 +656,12 @@ _iodbcdm_SetConnectOption (
 	  return SQL_ERROR;
 	}
 
-      STRCPY (pdbc->current_qualifier, vParam);
+      if (waMode != 'W')
+        STRCPY ((char *)pdbc->current_qualifier, vParam);
+      else
+        WCSCPY ((wchar_t *)pdbc->current_qualifier, vParam);
+
+      pdbc->current_qualifier_WA = waMode;
       break;
 
     case SQL_LOGIN_TIMEOUT:
@@ -565,7 +685,7 @@ _iodbcdm_SetConnectOption (
       break;
 
     default:
-      /* Since we didn't save the option value for delaied
+      /* Since we didn't save the option value for delayed
        * setting, we should raise an error here.
        */
       break;
@@ -574,20 +694,52 @@ _iodbcdm_SetConnectOption (
   return retcode;
 }
 
+
 SQLRETURN SQL_API
 SQLSetConnectOption (
-    SQLHDBC hdbc,
-    SQLUSMALLINT fOption,
-    SQLUINTEGER vParam)
+  SQLHDBC		  hdbc,
+  SQLUSMALLINT		  fOption,
+  SQLUINTEGER		  vParam)
 {
-  CONN (pdbc, hdbc);
-  SQLRETURN retcode;
+  ENTER_HDBC (hdbc, 1,
+    trace_SQLSetConnectOption (TRACE_ENTER, hdbc, fOption, vParam));
 
-  ENTER_HDBC (pdbc);
+  retcode = _iodbcdm_SetConnectOption (hdbc, fOption, vParam, 'A');
 
-  retcode = _iodbcdm_SetConnectOption (hdbc, fOption, vParam);
+  LEAVE_HDBC (hdbc, 1,
+    trace_SQLSetConnectOption (TRACE_LEAVE, hdbc, fOption, vParam));
+}
 
-  LEAVE_HDBC (pdbc, retcode);
+
+SQLRETURN SQL_API
+SQLSetConnectOptionA (
+  SQLHDBC		  hdbc,
+  SQLUSMALLINT		  fOption,
+  SQLUINTEGER		  vParam)
+{
+  ENTER_HDBC (hdbc, 1,
+    trace_SQLSetConnectOption (TRACE_ENTER, hdbc, fOption, vParam));
+
+  retcode = _iodbcdm_SetConnectOption (hdbc, fOption, vParam, 'A');
+
+  LEAVE_HDBC (hdbc, 1,
+    trace_SQLSetConnectOption (TRACE_LEAVE, hdbc, fOption, vParam));
+}
+
+
+SQLRETURN SQL_API
+SQLSetConnectOptionW (
+  SQLHDBC		  hdbc,
+  SQLUSMALLINT		  fOption,
+  SQLUINTEGER		  vParam)
+{
+  ENTER_HDBC (hdbc, 1,
+    trace_SQLSetConnectOptionW (TRACE_ENTER, hdbc, fOption, vParam));
+
+  retcode = _iodbcdm_SetConnectOption (hdbc, fOption, vParam, 'W');
+
+  LEAVE_HDBC (hdbc, 1,
+    trace_SQLSetConnectOptionW (TRACE_LEAVE, hdbc, fOption, vParam));
 }
 
 
@@ -595,12 +747,14 @@ SQLRETURN SQL_API
 _iodbcdm_GetConnectOption (
     SQLHDBC hdbc,
     SQLUSMALLINT fOption,
-    SQLPOINTER pvParam)
+    SQLPOINTER pvParam,
+    SQLCHAR waMode)
 {
   CONN (pdbc, hdbc);
-  int sqlstat = en_00000;
+  ENV_t FAR *penv = pdbc->henv;
   HPROC hproc = SQL_NULL_HPROC;
-  SQLRETURN retcode;
+  int sqlstat = en_00000;
+  SQLRETURN retcode = SQL_SUCCESS;
 
 #if (ODBCVER < 0x0300)
   /* check option */
@@ -646,7 +800,6 @@ _iodbcdm_GetConnectOption (
   if (sqlstat != en_00000)
     {
       PUSHSQLERR (pdbc->herr, sqlstat);
-
       return SQL_ERROR;
     }
 
@@ -659,7 +812,7 @@ _iodbcdm_GetConnectOption (
   if (fOption == SQL_OPT_TRACE)
 #endif
     {
-      if (pdbc->trace)
+      if (ODBCSharedTraceFlag)
 	*((UDWORD *) pvParam) = (UDWORD) SQL_OPT_TRACE_ON;
       else
 	*((UDWORD *) pvParam) = (UDWORD) SQL_OPT_TRACE_OFF;
@@ -673,7 +826,20 @@ _iodbcdm_GetConnectOption (
   if (fOption == SQL_OPT_TRACEFILE)
 #endif
     {
-      STRCPY (pvParam, pdbc->tfile);
+      char *fname;
+      
+      fname = trace_get_filename ();	/* UTF8 format */
+
+      if (waMode != 'W')
+        {
+           STRCPY (pvParam, fname);
+        }
+      else
+        {
+           dm_strcpy_A2W (pvParam, (SQLCHAR *)fname);
+        }
+
+      free (fname);
 
       return SQL_SUCCESS;
     }
@@ -682,9 +848,62 @@ _iodbcdm_GetConnectOption (
     /* if already connected, we will invoke driver's function */
     {
 
+      void * _Param = NULL;
+      void * paramOut = pvParam;
+      int procid = 0;
+
+      switch (fOption)
+        {
+        case SQL_CURRENT_QUALIFIER:
+        case SQL_TRANSLATE_DLL:
+
+          if ((penv->unicode_driver && waMode != 'W') 
+              || (!penv->unicode_driver && waMode == 'W'))
+            {
+              if (waMode != 'W')
+                {
+                /* ansi=>unicode*/
+                  if ((_Param = malloc(SQL_MAX_OPTION_STRING_LENGTH * sizeof(wchar_t))) == NULL)
+	            {
+                      PUSHSQLERR (pdbc->herr, en_HY001);
+                      return SQL_ERROR;
+                    }
+                }
+              else
+                {
+                /* unicode=>ansi*/
+                  if ((_Param = malloc(SQL_MAX_OPTION_STRING_LENGTH)) == NULL)
+    	            {
+                      PUSHSQLERR (pdbc->herr, en_HY001);
+                      return SQL_ERROR;
+                    }
+                }
+              paramOut = _Param;
+            }
+          break;
+        }
+
 #if (ODBCVER >= 0x0300)
 
-      hproc = _iodbcdm_getproc (pdbc, en_GetConnectAttr);
+     if (penv->unicode_driver)
+       {
+         /* SQL_XXX_W */
+         if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectAttrW))
+             != SQL_NULL_HPROC)
+           procid = en_GetConnectAttrW;
+       }
+     else
+       {
+         /* SQL_XXX */
+         /* SQL_XXX_A */
+         if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectAttr))
+             != SQL_NULL_HPROC)
+           procid = en_GetConnectAttr;
+         else
+         if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectAttrA))
+             != SQL_NULL_HPROC)
+           procid = en_GetConnectAttrA;
+       }
 
       if (hproc != SQL_NULL_HPROC)
 	{
@@ -699,8 +918,8 @@ _iodbcdm_GetConnectOption (
 	    case SQL_ATTR_QUIET_MODE:
 	    case SQL_ATTR_TRANSLATE_OPTION:
 	    case SQL_ATTR_TXN_ISOLATION:
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_GetConnectAttr,
-		  (pdbc->dhdbc, fOption, pvParam, 0, NULL));
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+		  (pdbc->dhdbc, fOption, paramOut, 0, NULL));
 	      break;
 
 	      /* ODBC3 defined options */
@@ -710,30 +929,75 @@ _iodbcdm_GetConnectOption (
 	    case SQL_ATTR_CONNECTION_TIMEOUT:
 	    case SQL_ATTR_METADATA_ID:
 	      PUSHSQLERR (pdbc->herr, en_IM001);
+	      MEM_FREE(_Param);
 	      return SQL_ERROR;
 
 	    default:		/* string & driver defined */
 
-	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_GetConnectAttr,
-		  (pdbc->dhdbc, fOption, pvParam, SQL_MAX_OPTION_STRING_LENGTH, NULL));
+	      CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+		  (pdbc->dhdbc, fOption, paramOut, SQL_MAX_OPTION_STRING_LENGTH, NULL));
+
 	    }
 	}
       else
 #endif
 	{
-	  hproc = _iodbcdm_getproc (pdbc, en_GetConnectOption);
+          if (penv->unicode_driver)
+            {
+             /* SQL_XXX_W */
+             if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectOptionW))
+                 != SQL_NULL_HPROC)
+               procid = en_GetConnectOptionW;
+            }
+          else
+            {
+             /* SQL_XXX */
+             /* SQL_XXX_A */
+             if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectOption))
+                != SQL_NULL_HPROC)
+               procid = en_GetConnectOption;
+             else
+             if ((hproc = _iodbcdm_getproc (pdbc, en_GetConnectOptionA))
+                != SQL_NULL_HPROC)
+               procid = en_GetConnectOptionA;
+            }
 
 	  if (hproc == SQL_NULL_HPROC)
 	    {
 	      PUSHSQLERR (pdbc->herr, en_IM001);
-
+              MEM_FREE(_Param);
 	      return SQL_ERROR;
 	    }
 
-	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_GetConnectOption,
-	      (pdbc->dhdbc, fOption, pvParam));
+	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, procid,
+	      (pdbc->dhdbc, fOption, paramOut));
 	}
 
+      if (pvParam
+          && (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
+          &&  ((penv->unicode_driver && waMode != 'W') 
+              || (!penv->unicode_driver && waMode == 'W')))
+        {
+          switch (fOption)
+            {
+            case SQL_ATTR_TRACEFILE:
+            case SQL_CURRENT_QUALIFIER:
+            case SQL_TRANSLATE_DLL:
+              if (waMode != 'W')
+                {
+                 /* ansi<=unicode*/
+                  dm_StrCopyOut2_W2A (paramOut, pvParam, SQL_MAX_OPTION_STRING_LENGTH, NULL);
+                }
+              else
+                {
+                 /* unicode<=ansi*/
+                  dm_StrCopyOut2_A2W (paramOut, pvParam, SQL_MAX_OPTION_STRING_LENGTH, NULL);
+                }
+              break;
+            }
+        }
+      
+      MEM_FREE(_Param);
       return retcode;
     }
 
@@ -763,20 +1027,51 @@ _iodbcdm_GetConnectOption (
   return SQL_SUCCESS;
 }
 
+
 SQLRETURN SQL_API
-SQLGetConnectOption (
-    SQLHDBC hdbc,
+SQLGetConnectOption (SQLHDBC hdbc,
     SQLUSMALLINT fOption,
     SQLPOINTER pvParam)
 {
-  CONN (pdbc, hdbc);
-  SQLRETURN retcode;
+  ENTER_HDBC (hdbc, 0,
+    trace_SQLGetConnectOption (TRACE_ENTER, hdbc, fOption, pvParam));
 
-  ENTER_HDBC (pdbc);
+  retcode = _iodbcdm_GetConnectOption (hdbc, fOption, pvParam, 'A');
 
-  retcode = _iodbcdm_GetConnectOption (hdbc, fOption, pvParam);
+  LEAVE_HDBC (hdbc, 0,
+    trace_SQLGetConnectOption (TRACE_LEAVE, hdbc, fOption, pvParam));
+}
 
-  LEAVE_HDBC (pdbc, retcode);
+
+SQLRETURN SQL_API
+SQLGetConnectOptionA (
+  SQLHDBC		  hdbc,
+  SQLUSMALLINT		  fOption,
+  SQLPOINTER		  pvParam)
+{
+  ENTER_HDBC (hdbc, 1,
+    trace_SQLGetConnectOption (TRACE_ENTER, hdbc, fOption, pvParam));
+
+  retcode = _iodbcdm_GetConnectOption (hdbc, fOption, pvParam, 'A');
+
+  LEAVE_HDBC (hdbc, 1,
+    trace_SQLGetConnectOption (TRACE_LEAVE, hdbc, fOption, pvParam));
+}
+
+
+SQLRETURN SQL_API
+SQLGetConnectOptionW (
+  SQLHDBC		  hdbc,
+  SQLUSMALLINT		  fOption,
+  SQLPOINTER		  pvParam)
+{
+  ENTER_HDBC (hdbc, 1,
+    trace_SQLGetConnectOptionW (TRACE_ENTER, hdbc, fOption, pvParam));
+
+  retcode = _iodbcdm_GetConnectOption (hdbc, fOption, pvParam, 'W');
+
+  LEAVE_HDBC (hdbc, 1,
+    trace_SQLGetConnectOptionW (TRACE_LEAVE, hdbc, fOption, pvParam));
 }
 
 
@@ -831,7 +1126,7 @@ _iodbcdm_transact (
       hproc = _iodbcdm_getproc (pdbc, en_EndTran);
       if (hproc != SQL_NULL_HPROC)
 	{
-	  CALL_DRIVER (pdbc, pdbc, retcode, hproc, en_EndTran,
+	  CALL_DRIVER (hdbc, pdbc, retcode, hproc, en_EndTran,
 	      (SQL_HANDLE_DBC, pdbc->dhdbc, fType));
 	}
       else
@@ -869,6 +1164,7 @@ _iodbcdm_transact (
 	    }
 	  break;
 
+	case en_stmt_executed_with_info:
 	case en_stmt_executed:
 	case en_stmt_cursoropen:
 	case en_stmt_fetched:
@@ -916,8 +1212,8 @@ _iodbcdm_transact (
 }
 
 
-SQLRETURN SQL_API
-SQLTransact (
+SQLRETURN 
+SQLTransact_Internal (
     SQLHENV henv,
     SQLHDBC hdbc,
     SQLUSMALLINT fType)
@@ -927,7 +1223,6 @@ SQLTransact (
   HERR herr;
   SQLRETURN retcode = SQL_SUCCESS;
 
-  ODBC_LOCK ();
   if (IS_VALID_HDBC (pdbc))
     {
       CLEAR_ERRORS (pdbc);
@@ -940,16 +1235,14 @@ SQLTransact (
     }
   else
     {
-      ODBC_UNLOCK ();
       return SQL_INVALID_HANDLE;
     }
 
   /* check argument */
   if (fType != SQL_COMMIT && fType != SQL_ROLLBACK)
     {
-      SQLHENV handle = IS_VALID_HDBC (pdbc) ? ((SQLHENV) pdbc) : genv;
+/*      SQLHENV handle = IS_VALID_HDBC (pdbc) ? ((SQLHENV) pdbc) : genv;*/
       PUSHSQLERR (herr, en_S1012);
-      ODBC_UNLOCK ();
       return SQL_ERROR;
     }
 
@@ -967,11 +1260,29 @@ SQLTransact (
 
   if (retcode != SQL_SUCCESS && retcode != SQL_SUCCESS_WITH_INFO)
     {
-      ODBC_UNLOCK ();
       /* fail on one of the connection */
       return SQL_ERROR;
     }
 
+  return retcode;
+}
+
+
+SQLRETURN SQL_API
+SQLTransact (
+  SQLHENV		  henv,
+  SQLHDBC		  hdbc,
+  SQLUSMALLINT		  fType)
+{
+  SQLRETURN retcode = SQL_SUCCESS;
+
+  ODBC_LOCK ();
+  TRACE (trace_SQLTransact (TRACE_ENTER, henv, hdbc, fType));
+
+  retcode = SQLTransact_Internal (henv, hdbc, fType);
+
+  TRACE (trace_SQLTransact (TRACE_LEAVE, henv, hdbc, fType));
   ODBC_UNLOCK ();
+
   return retcode;
 }
