@@ -1,14 +1,32 @@
 /*
  *  prepare.c
  *
- *  $Id: prepare.c,v 1.11 2001/06/04 14:01:25 source Exp $
+ *  $Id: prepare.c,v 1.15 2003/09/02 14:44:22 source Exp $
  *
  *  Prepare a query
  *
  *  The iODBC driver manager.
  *  
  *  Copyright (C) 1995 by Ke Jin <kejin@empress.com> 
+ *  Copyright (C) 1996-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
+ *  This software is released under the terms of either of the following
+ *  licenses:
+ *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -22,12 +40,44 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <iodbc.h>
 
 #include <sql.h>
 #include <sqlext.h>
+#include <sqlucode.h>
+
+#include <unicode.h>
 
 #include <dlproc.h>
 
@@ -39,18 +89,19 @@
 #include <itrace.h>
 
 SQLRETURN SQL_API
-SQLPrepare (
+SQLPrepare_Internal (
     SQLHSTMT hstmt,
-    SQLCHAR FAR * szSqlStr,
-    SQLINTEGER cbSqlStr)
+    SQLPOINTER szSqlStr,
+    SQLINTEGER cbSqlStr,
+    SQLCHAR waMode)
 {
   STMT (pstmt, hstmt);
-
+  CONN (pdbc, pstmt->hdbc);
+  ENV_t FAR *penv = pdbc->henv;
   HPROC hproc = SQL_NULL_HPROC;
   SQLRETURN retcode = SQL_SUCCESS;
   int sqlstat = en_00000;
-
-  ENTER_STMT (pstmt);
+  void * _SqlStr = NULL;
 
   /* check state */
   if (pstmt->asyn_on == en_NullProc)
@@ -83,33 +134,55 @@ SQLPrepare (
     {
       PUSHSQLERR (pstmt->herr, sqlstat);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   if (szSqlStr == NULL)
     {
       PUSHSQLERR (pstmt->herr, en_S1009);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   if (cbSqlStr < 0 && cbSqlStr != SQL_NTS)
     {
       PUSHSQLERR (pstmt->herr, en_S1090);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
-  hproc = _iodbcdm_getproc (pstmt->hdbc, en_Prepare);
+  if ((penv->unicode_driver && waMode != 'W')
+      || (!penv->unicode_driver && waMode == 'W'))
+    {
+      if (waMode != 'W')
+        {
+        /* ansi=>unicode*/
+          _SqlStr = _iodbcdm_conv_param_A2W(pstmt, 0, szSqlStr, cbSqlStr);
+        }
+      else
+        {
+        /* unicode=>ansi*/
+          _SqlStr = _iodbcdm_conv_param_W2A(pstmt, 0, szSqlStr, cbSqlStr);
+        }
+      szSqlStr = _SqlStr;
+      cbSqlStr = SQL_NTS;
+    }
+
+  CALL_UDRIVER(pstmt->hdbc, pstmt, retcode, hproc, penv->unicode_driver, 
+    en_Prepare, (
+       pstmt->dhstmt,
+       szSqlStr,
+       cbSqlStr));
 
   if (hproc == SQL_NULL_HPROC)
     {
+      _iodbcdm_FreeStmtParams(pstmt);
       PUSHSQLERR (pstmt->herr, en_IM001);
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
-  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_Prepare,
-      (pstmt->dhstmt, szSqlStr, cbSqlStr));
+  if (retcode != SQL_STILL_EXECUTING)
+    _iodbcdm_FreeStmtParams(pstmt);
 
   /* stmt state transition */
   if (pstmt->asyn_on == en_Prepare)
@@ -120,11 +193,11 @@ SQLPrepare (
 	case SQL_SUCCESS_WITH_INFO:
 	case SQL_ERROR:
 	  pstmt->asyn_on = en_NullProc;
-	   LEAVE_STMT (pstmt, retcode);
+	   return retcode;
 
 	case SQL_STILL_EXECUTING:
 	default:
-	   LEAVE_STMT (pstmt, retcode);
+	   return retcode;
 	}
     }
 
@@ -144,6 +217,7 @@ SQLPrepare (
       switch (pstmt->state)
 	{
 	case en_stmt_prepared:
+	case en_stmt_executed_with_info:
 	case en_stmt_executed:
 	  pstmt->state = en_stmt_allocated;
 	  pstmt->prep_state = 0;
@@ -157,36 +231,82 @@ SQLPrepare (
       break;
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLSetCursorName (
+SQLPrepare (SQLHSTMT hstmt,
+    SQLCHAR FAR * szSqlStr,
+    SQLINTEGER cbSqlStr)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLPrepare (TRACE_ENTER, hstmt, szSqlStr, cbSqlStr));
+
+  retcode = SQLPrepare_Internal(hstmt, szSqlStr, cbSqlStr, 'A');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLPrepare (TRACE_LEAVE, hstmt, szSqlStr, cbSqlStr));
+}
+
+
+SQLRETURN SQL_API
+SQLPrepareA (SQLHSTMT hstmt,
+    SQLCHAR FAR * szSqlStr,
+    SQLINTEGER cbSqlStr)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLPrepare (TRACE_ENTER, hstmt, szSqlStr, cbSqlStr));
+
+  retcode = SQLPrepare_Internal(hstmt, szSqlStr, cbSqlStr, 'A');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLPrepare (TRACE_LEAVE, hstmt, szSqlStr, cbSqlStr));
+}
+
+
+SQLRETURN SQL_API
+SQLPrepareW (SQLHSTMT hstmt,
+    SQLWCHAR FAR * szSqlStr,
+    SQLINTEGER cbSqlStr)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLPrepareW (TRACE_ENTER, hstmt, szSqlStr, cbSqlStr));
+
+  retcode = SQLPrepare_Internal(hstmt, szSqlStr, cbSqlStr, 'W');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLPrepareW (TRACE_LEAVE, hstmt, szSqlStr, cbSqlStr));
+}
+
+
+SQLRETURN SQL_API
+SQLSetCursorName_Internal (
     SQLHSTMT hstmt,
-    SQLCHAR FAR * szCursor,
-    SQLSMALLINT cbCursor)
+    SQLPOINTER szCursor,
+    SQLSMALLINT cbCursor,
+    SQLCHAR waMode)
 {
   STMT (pstmt, hstmt);
+  CONN (pdbc, pstmt->hdbc);
+  ENV_t FAR *penv = pdbc->henv;
   HPROC hproc = SQL_NULL_HPROC;
-
   SQLRETURN retcode = SQL_SUCCESS;
   int sqlstat = en_00000;
-
-  ENTER_STMT (pstmt);
+  void * _Cursor = NULL;
 
   if (szCursor == NULL)
     {
       PUSHSQLERR (pstmt->herr, en_S1009);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   if (cbCursor < 0 && cbCursor != SQL_NTS)
     {
       PUSHSQLERR (pstmt->herr, en_S1090);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* check state */
@@ -198,6 +318,7 @@ SQLSetCursorName (
     {
       switch (pstmt->state)
 	{
+	case en_stmt_executed_with_info:
 	case en_stmt_executed:
 	case en_stmt_cursoropen:
 	case en_stmt_fetched:
@@ -220,33 +341,98 @@ SQLSetCursorName (
     {
       PUSHSQLERR (pstmt->herr, sqlstat);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
-  hproc = _iodbcdm_getproc (pstmt->hdbc, en_SetCursorName);
+  if ((penv->unicode_driver && waMode != 'W')
+      || (!penv->unicode_driver && waMode == 'W'))
+    {
+      if (waMode != 'W')
+        {
+        /* ansi=>unicode*/
+          _Cursor = dm_SQL_A2W(szCursor, cbCursor);
+        }
+      else
+        {
+        /* unicode=>ansi*/
+          _Cursor = dm_SQL_W2A(szCursor, cbCursor);
+        }
+      szCursor = _Cursor;
+      cbCursor = SQL_NTS;
+    }
+
+  CALL_UDRIVER(pstmt->hdbc, pstmt, retcode, hproc, penv->unicode_driver, 
+    en_SetCursorName, (
+       pstmt->dhstmt,
+       szCursor,
+       cbCursor));
+
+  MEM_FREE(_Cursor);
 
   if (hproc == SQL_NULL_HPROC)
     {
       PUSHSQLERR (pstmt->herr, en_IM001);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
-  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_SetCursorName,
-      (pstmt->dhstmt, szCursor, cbCursor));
-
   if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO)
     {
       pstmt->cursor_state = en_stmt_cursor_named;
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
+SQLRETURN SQL_API
+SQLSetCursorName (SQLHSTMT hstmt,
+    SQLCHAR FAR * szCursor,
+    SQLSMALLINT cbCursor)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLSetCursorName (TRACE_ENTER, hstmt, szCursor, cbCursor));
+
+  retcode = SQLSetCursorName_Internal(hstmt, szCursor, cbCursor, 'A');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLSetCursorName (TRACE_LEAVE, hstmt, szCursor, cbCursor));
+}
+
 
 SQLRETURN SQL_API
-SQLBindParameter (
+SQLSetCursorNameA (SQLHSTMT hstmt,
+    SQLCHAR FAR * szCursor,
+    SQLSMALLINT cbCursor)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLSetCursorName (TRACE_ENTER, hstmt, szCursor, cbCursor));
+
+  retcode = SQLSetCursorName_Internal(hstmt, szCursor, cbCursor, 'A');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLSetCursorName (TRACE_LEAVE, hstmt, szCursor, cbCursor));
+}
+
+
+SQLRETURN SQL_API
+SQLSetCursorNameW (
+  SQLHSTMT		  hstmt,
+  SQLWCHAR FAR		* szCursor,
+  SQLSMALLINT		  cbCursor)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLSetCursorNameW (TRACE_ENTER, hstmt, szCursor, cbCursor));
+
+  retcode = SQLSetCursorName_Internal(hstmt, szCursor, cbCursor, 'W');
+
+  LEAVE_STMT (hstmt,
+    trace_SQLSetCursorNameW (TRACE_LEAVE, hstmt, szCursor, cbCursor));
+}
+
+
+static SQLRETURN
+SQLBindParameter_Internal (
     SQLHSTMT hstmt,
     SQLUSMALLINT ipar,
     SQLSMALLINT fParamType,
@@ -259,20 +445,22 @@ SQLBindParameter (
     SQLINTEGER FAR * pcbValue)
 {
   STMT (pstmt, hstmt);
+  CONN (pdbc, pstmt->hdbc);
+  ENV_t FAR *penv = pdbc->henv;
   HPROC hproc = SQL_NULL_HPROC;
+  SQLSMALLINT nCType;
+  SQLSMALLINT nSqlType;
 
   int sqlstat = en_00000;
   SQLRETURN retcode = SQL_SUCCESS;
 
-  ENTER_STMT (pstmt);
-
 #if (ODBCVER >= 0x0300)
   if (0)
 #else
   /* check param */
   if (fSqlType > SQL_TYPE_MAX ||
       (fSqlType < SQL_TYPE_MIN && fSqlType > SQL_TYPE_DRIVER_START))
-    /* Note: SQL_TYPE_DRIVER_START is a nagtive number 
+    /* Note: SQL_TYPE_DRIVER_START is a negative number 
      * So, we use ">" */
 #endif
     {
@@ -346,6 +534,7 @@ SQLBindParameter (
 	case SQL_C_TYPE_TIME:
 	case SQL_C_TYPE_TIMESTAMP:
 	case SQL_C_UBIGINT:
+	case SQL_C_WCHAR:
 #endif
 	  break;
 
@@ -359,7 +548,7 @@ SQLBindParameter (
     {
       PUSHSQLERR (pstmt->herr, sqlstat);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   /* check state */
@@ -370,6 +559,16 @@ SQLBindParameter (
       retcode = SQL_ERROR;
     }
 
+  /*
+   *  Convert C type to ODBC version of driver
+   */
+  nCType = _iodbcdm_map_c_type (fCType, penv->dodbc_ver);
+
+  /*
+   *  Convert SQL type to ODBC version of driver
+   */
+  nSqlType = _iodbcdm_map_sql_type (fSqlType, penv->dodbc_ver);
+
 #if (ODBCVER >=0x0300)
   if (fParamType == SQL_PARAM_INPUT)
     {
@@ -377,9 +576,9 @@ SQLBindParameter (
       if (hproc)
 	{
 	  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_BindParam,
-	      (pstmt->dhstmt, ipar, fCType, fSqlType, cbColDef,
+	      (pstmt->dhstmt, ipar, nCType, nSqlType, cbColDef,
 		  ibScale, rgbValue, pcbValue));
-	  LEAVE_STMT (pstmt, retcode);
+	  return retcode;
 	}
     }
 #endif
@@ -391,41 +590,68 @@ SQLBindParameter (
 
       PUSHSQLERR (pstmt->herr, en_IM001);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_BindParameter,
-      (pstmt->dhstmt, ipar, fParamType, fCType, fSqlType, cbColDef,
+      (pstmt->dhstmt, ipar, fParamType, nCType, nSqlType, cbColDef,
 	  ibScale, rgbValue, cbValueMax, pcbValue));
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLParamOptions (
+SQLBindParameter (
     SQLHSTMT hstmt,
-    SQLUINTEGER crow,
-    SQLUINTEGER FAR * pirow)
+    SQLUSMALLINT ipar,
+    SQLSMALLINT fParamType,
+    SQLSMALLINT fCType,
+    SQLSMALLINT fSqlType,
+    SQLUINTEGER cbColDef,
+    SQLSMALLINT ibScale,
+    SQLPOINTER rgbValue,
+    SQLINTEGER cbValueMax,
+    SQLINTEGER FAR * pcbValue)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLBindParameter (TRACE_ENTER,
+	hstmt, ipar, fParamType, fCType, fSqlType, cbColDef,
+	ibScale, rgbValue, cbValueMax, pcbValue));
+
+  retcode = SQLBindParameter_Internal (
+	hstmt, ipar, fParamType, fCType, fSqlType, cbColDef,
+	ibScale, rgbValue, cbValueMax, pcbValue);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLBindParameter (TRACE_LEAVE,
+	hstmt, ipar, fParamType, fCType, fSqlType, cbColDef,
+	ibScale, rgbValue, cbValueMax, pcbValue));
+}
+
+
+static SQLRETURN 
+SQLParamOptions_Internal (
+  SQLHSTMT		  hstmt,
+  SQLUINTEGER		  crow,
+  SQLUINTEGER FAR	* pirow)
 {
   STMT (pstmt, hstmt);
-  HPROC hproc;
+  HPROC hproc = SQL_NULL_HPROC;
   SQLRETURN retcode;
 
-  ENTER_STMT (pstmt);
-
   if (crow == (UDWORD) 0UL)
     {
       PUSHSQLERR (pstmt->herr, en_S1107);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   if (pstmt->state >= en_stmt_needdata || pstmt->asyn_on != en_NullProc)
     {
       PUSHSQLERR (pstmt->herr, en_S1010);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
 #if (ODBCVER >= 0x0300)
@@ -452,31 +678,45 @@ SQLParamOptions (
 	{
 	  PUSHSQLERR (pstmt->herr, en_IM001);
 
-	  LEAVE_STMT (pstmt, SQL_ERROR);
+	  return SQL_ERROR;
 	}
 
       CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_ParamOptions,
 	  (pstmt->dhstmt, crow, pirow));
     }
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
 }
 
 
 SQLRETURN SQL_API
-SQLSetScrollOptions (
-    SQLHSTMT hstmt,
-    SQLUSMALLINT fConcurrency,
-    SQLINTEGER crowKeyset,
-    SQLUSMALLINT crowRowset)
+SQLParamOptions(
+  SQLHSTMT		  hstmt,
+  SQLUINTEGER		  crow,
+  SQLUINTEGER FAR	* pirow)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLParamOptions (TRACE_ENTER, hstmt, crow, pirow));
+
+  retcode = SQLParamOptions_Internal (hstmt, crow, pirow);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLParamOptions (TRACE_LEAVE, hstmt, crow, pirow));
+}
+
+
+static SQLRETURN
+SQLSetScrollOptions_Internal (
+  SQLHSTMT		  hstmt,
+  SQLUSMALLINT		  fConcurrency,
+  SQLINTEGER		  crowKeyset,
+  SQLUSMALLINT		  crowRowset)
 {
   STMT (pstmt, hstmt);
-  HPROC hproc;
+  HPROC hproc = SQL_NULL_HPROC;
   int sqlstat = en_00000;
   SQLRETURN retcode;
 
-  ENTER_STMT (pstmt);
-
   for (;;)
     {
       if (crowRowset == (UWORD) 0)
@@ -531,7 +771,7 @@ SQLSetScrollOptions (
 	  if (hproc1 == SQL_NULL_HPROC)
 	    {
 	      PUSHSQLERR (pstmt->herr, en_IM001);
-	      LEAVE_STMT (pstmt, SQL_ERROR);
+	      return SQL_ERROR;
 	    }
 
 	  switch (crowKeyset)
@@ -561,7 +801,7 @@ SQLSetScrollOptions (
 	  retcode = SQLGetInfo (pstmt->hdbc, InfoType, &InfoValue, 0, NULL);
 	  if (retcode != SQL_SUCCESS)
 	    {
-	      LEAVE_STMT (pstmt, retcode);
+	      return retcode;
 	    }
 
 	  switch (fConcurrency)
@@ -570,7 +810,7 @@ SQLSetScrollOptions (
 	      if (!(InfoValue & SQL_CA2_READ_ONLY_CONCURRENCY))
 		{
 		  PUSHSQLERR (pstmt->herr, en_S1C00);
-		  LEAVE_STMT (pstmt, SQL_ERROR);
+		  return SQL_ERROR;
 		}
 	      break;
 
@@ -578,7 +818,7 @@ SQLSetScrollOptions (
 	      if (!(InfoValue & SQL_CA2_LOCK_CONCURRENCY))
 		{
 		  PUSHSQLERR (pstmt->herr, en_S1C00);
-		  LEAVE_STMT (pstmt, SQL_ERROR);
+		  return SQL_ERROR;
 		}
 	      break;
 
@@ -586,7 +826,7 @@ SQLSetScrollOptions (
 	      if (!(InfoValue & SQL_CA2_OPT_ROWVER_CONCURRENCY))
 		{
 		  PUSHSQLERR (pstmt->herr, en_S1C00);
-		  LEAVE_STMT (pstmt, SQL_ERROR);
+		  return SQL_ERROR;
 		}
 	      break;
 
@@ -594,7 +834,7 @@ SQLSetScrollOptions (
 	      if (!(InfoValue & SQL_CA2_OPT_VALUES_CONCURRENCY))
 		{
 		  PUSHSQLERR (pstmt->herr, en_S1C00);
-		  LEAVE_STMT (pstmt, SQL_ERROR);
+		  return SQL_ERROR;
 		}
 	      break;
 	    }
@@ -602,13 +842,13 @@ SQLSetScrollOptions (
 	      (pstmt->dhstmt, SQL_ATTR_CURSOR_TYPE, Value, 0));
 
 	  if (retcode != SQL_SUCCESS)
-	    LEAVE_STMT (pstmt, retcode);
+	    return retcode;
 
 	  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc1, en_SetStmtAttr,
 	      (pstmt->dhstmt, SQL_ATTR_CONCURRENCY, fConcurrency, 0));
 
 	  if (retcode != SQL_SUCCESS)
-	    LEAVE_STMT (pstmt, retcode);
+	    return retcode;
 
 	  if (crowKeyset > 0)
 	    {
@@ -616,14 +856,14 @@ SQLSetScrollOptions (
 		  (pstmt->dhstmt, SQL_ATTR_KEYSET_SIZE, crowKeyset, 0));
 
 	      if (retcode != SQL_SUCCESS)
-		LEAVE_STMT (pstmt, retcode);
+		return retcode;
 	    }
 
 	  CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc1, en_SetStmtAttr,
 	      (pstmt->dhstmt, SQL_ROWSET_SIZE, crowRowset, 0));
 
 	  if (retcode != SQL_SUCCESS)
-	    LEAVE_STMT (pstmt, retcode);
+	    return retcode;
 #else
 	  sqlstat = en_IM001;
 	  break;
@@ -639,13 +879,42 @@ SQLSetScrollOptions (
     {
       PUSHSQLERR (pstmt->herr, sqlstat);
 
-      LEAVE_STMT (pstmt, SQL_ERROR);
+      return SQL_ERROR;
     }
 
   CALL_DRIVER (pstmt->hdbc, pstmt, retcode, hproc, en_SetScrollOptions,
       (pstmt->dhstmt, fConcurrency, crowKeyset, crowRowset));
 
-  LEAVE_STMT (pstmt, retcode);
+  return retcode;
+}
+
+
+SQLRETURN SQL_API
+SQLSetScrollOptions (
+  SQLHSTMT		  hstmt,
+  SQLUSMALLINT		  fConcurrency,
+  SQLINTEGER		  crowKeyset,
+  SQLUSMALLINT		  crowRowset)
+{
+  ENTER_STMT (hstmt,
+    trace_SQLSetScrollOptions (TRACE_ENTER,
+    	hstmt,
+	fConcurrency,
+	crowKeyset,
+	crowRowset));
+
+  retcode = SQLSetScrollOptions_Internal (
+    	hstmt,
+	fConcurrency,
+	crowKeyset,
+	crowRowset);
+
+  LEAVE_STMT (hstmt,
+    trace_SQLSetScrollOptions (TRACE_LEAVE,
+    	hstmt,
+	fConcurrency,
+	crowKeyset,
+	crowRowset));
 }
 
 
