@@ -1,12 +1,29 @@
 /*
  *  administrator.c
  *
- *  $Id: administrator.c,v 1.2 2001/06/11 12:36:50 source Exp $
+ *  $Id: administrator.c,v 1.7 2003/09/07 12:59:14 source Exp $
  *
  *  The iODBC driver manager.
+ *  
+ *  Copyright (C) 1999-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
- *  Copyright (C) 2001 by OpenLink Software <iodbc@openlinksw.com>
+ *  This software is released under the terms of either of the following
+ *  licenses:
  *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -20,8 +37,38 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#include <iodbc.h>
 #include <dlfcn.h>
 #include <sys/stat.h>
 #include <unistd.h>
@@ -29,17 +76,15 @@
 #include "../gui.h"
 #include "odbc4.xpm"
 
-#if defined(HAVE_DLADDR)
-#if defined(linux)
+#if !defined(HAVE_DL_INFO)
 typedef struct
 {
-  __const char *dli_fname;	/* File name of defining object.  */
+  const char *dli_fname;	/* File name of defining object.  */
   void *dli_fbase;		/* Load address of that object.  */
-  __const char *dli_sname;	/* Name of nearest symbol.  */
+  const char *dli_sname;	/* Name of nearest symbol.  */
   void *dli_saddr;		/* Exact value of nearest symbol.  */
 } Dl_info;
-#endif
-#endif /* HAVE_DLADDR */
+#endif /* HAVE_DL_INFO */
 
 static char *szDriverButtons[] = {
   "_Add a driver",
@@ -52,12 +97,16 @@ static char *szCpLabels[] = {
   "Pool timeout (seconds)"
 };
 
-static char *szComponentNames[] = {
-  "libiodbc.so",
-  "libiodbcadm.so",
-  "libiodbcinst.so",
-  "libdrvproxy.so",
-  "libtranslator.so"
+static struct {
+  char *lib_name;
+  char *lib_desc;
+  char *lib_ver_sym;
+} iODBC_Components[]  = {
+  { "libiodbc.so",	"iODBC Driver Manager", 	"iodbc_version"},
+  { "libiodbcadm.so", 	"iODBC Administrator",		"iodbcadm_version"},
+  { "libiodbcinst.so",	"iODBC Installer",		"iodbcinst_version"},
+  { "libdrvproxy.so", 	"iODBC Driver Setup Proxy",	"iodbcproxy_version"},
+  { "libtranslator.so", "iODBC Translation Manager", 	"iodbctrans_version"}
 };
 
 
@@ -75,40 +124,39 @@ addcomponents_to_list (GtkWidget *widget)
 
   if (!GTK_IS_CLIST (widget))
     return;
+
   gtk_clist_clear (GTK_CLIST (widget));
 
-  for (i = 0; i < sizeof (szComponentNames) / sizeof (char *); i++)
+  for (i = 0; i < sizeof (iODBC_Components) / sizeof (iODBC_Components[0]); i++)
     {
-      if ((handle = dlopen (szComponentNames[i], RTLD_LAZY)))
-	{
-	  /* Take the name of the library */
-	  if ((proc = dlsym (handle, "libname")))
-	    {
-	      data[0] = *(char **) proc;
-	      data[1] = *(char **) dlsym (handle, "version");
-	      data[2] = szComponentNames[i];
+      /*
+       *  Collect basic info on the components
+       */
+      data[0] = iODBC_Components[i].lib_desc;
+      data[1] = VERSION;
+      data[2] = iODBC_Components[i].lib_name;
+      data[3] = "";				/* Modification Date */
+      data[4] = "";				/* Size */
 
-	      /* Take some information about the library */
+      if ((handle = dlopen (iODBC_Components[i].lib_name, RTLD_LAZY)))
+	{
+	  /* Find out the version of the library */
+	  if ((proc = dlsym (handle, iODBC_Components[i].lib_ver_sym)))
+	    data[1] = *(char **) proc;
+
+	  /* Check the size and modification date of the library */
 #ifdef HAVE_DLADDR
-	      dladdr (proc, &info);
-	      if (!stat (info.dli_fname, &_stat))
-		{
-		  sprintf (_size, "%d Kb", _stat.st_size / 1024);
-		  sprintf (_date, "%s", ctime (&_stat.st_mtime));
-		  data[3] = _date;
-		  data[4] = _size;
-		}
-	      else
-		{
-		  data[3] = "";
-		  data[4] = "";
-		}
-#else
-	      data[3] = "";
-	      data[4] = "";
-#endif
-	      gtk_clist_append (GTK_CLIST (widget), data);
+	  dladdr (proc, &info);	
+	  if (!stat (info.dli_fname, &_stat))
+	    {
+	      sprintf (_size, "%d Kb", _stat.st_size / 1024);
+	      sprintf (_date, "%s", ctime (&_stat.st_mtime));
+	      _date[strlen (_date) - 1] = '\0';	/* remove last \n */
+	      data[3] = _date;
+	      data[4] = _size;
 	    }
+#endif
+	  gtk_clist_append (GTK_CLIST (widget), data);
 
 	  dlclose (handle);
 	}
@@ -221,6 +269,22 @@ addconnectionpool_to_list (GtkWidget *widget)
 }
 
 
+static void
+admin_apply_tracing (TTRACING *tracing_t)
+{
+  char tokenstr[1024];
+  
+  /* Write keywords for tracing in the ini file */
+  SQLWritePrivateProfileString ("ODBC", "Trace",
+    (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (tracing_t->allthetime_rb)) == TRUE
+	 || gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (tracing_t->onetime_rb)) == TRUE) ? "1" : "0", NULL);
+  SQLWritePrivateProfileString ("ODBC", "TraceAutoStop",
+    gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (tracing_t->onetime_rb)) ? "1" : "0", NULL);
+  SQLWritePrivateProfileString ("ODBC", "TraceFile",
+	    gtk_entry_get_text (GTK_ENTRY (tracing_t->logfile_entry)), NULL);
+}
+
+
 static void
 admin_switch_page (GtkNotebook *notebook, GtkNotebookPage *page,
     gint page_num, void **inparams)
@@ -268,7 +332,7 @@ admin_switch_page (GtkNotebook *notebook, GtkNotebookPage *page,
     case 2:
       if (driverchoose_t)
 	{
-	  adddrivers_to_list (driverchoose_t->driverlist, FALSE);
+	  adddrivers_to_list (driverchoose_t->driverlist, driverchoose_t->mainwnd);
 	  gtk_widget_set_sensitive (driverchoose_t->b_remove, FALSE);
 	  gtk_widget_set_sensitive (driverchoose_t->b_configure, FALSE);
 	}
@@ -372,18 +436,7 @@ static void
 tracing_start_clicked (GtkWidget *widget, TTRACING *tracing_t)
 {
   if (tracing_t)
-    {
-      /* Write keywords for tracing in the ini file */
-      SQLWritePrivateProfileString ("ODBC", "Trace",
-	  (GTK_TOGGLE_BUTTON (tracing_t->donttrace_rb)->active) ? "0" : "1",
-	  NULL);
-      SQLWritePrivateProfileString ("ODBC", "TraceAutoStop",
-	  (GTK_TOGGLE_BUTTON (tracing_t->allthetime_rb)->active) ? "0" : "1",
-	  NULL);
-      if (STRLEN (gtk_entry_get_text (GTK_ENTRY (tracing_t->logfile_entry))))
-	SQLWritePrivateProfileString ("ODBC", "TraceFile",
-	    gtk_entry_get_text (GTK_ENTRY (tracing_t->logfile_entry)), NULL);
-    }
+    admin_apply_tracing (tracing_t);
 }
 
 
@@ -537,7 +590,7 @@ driver_add_clicked (GtkWidget *widget, TDRIVERCHOOSER *choose_t)
 	  free (cstr);
 	}
 
-      adddrivers_to_list (choose_t->driverlist, FALSE);
+      adddrivers_to_list (choose_t->driverlist, choose_t->mainwnd);
 
     done:
       if (GTK_CLIST (choose_t->driverlist)->selection == NULL)
@@ -582,7 +635,7 @@ driver_remove_clicked (GtkWidget *widget, TDRIVERCHOOSER *choose_t)
 	      goto done;
 	    }
 
-	  adddrivers_to_list (choose_t->driverlist, FALSE);
+	  adddrivers_to_list (choose_t->driverlist, choose_t->mainwnd);
 	}
 
     done:
@@ -660,7 +713,7 @@ driver_configure_clicked (GtkWidget *widget, TDRIVERCHOOSER *choose_t)
 	      free (cstr);
 	    }
 
-	  adddrivers_to_list (choose_t->driverlist, FALSE);
+	  adddrivers_to_list (choose_t->driverlist, choose_t->mainwnd);
 	}
 
     done:
@@ -712,20 +765,7 @@ admin_ok_clicked (GtkWidget *widget, void **inparams)
   if (tracing_t)
     {
       if (tracing_t->changed)
-	{
-	  /* Write keywords for tracing in the ini file */
-	  SQLWritePrivateProfileString ("ODBC", "Trace",
-	      (GTK_TOGGLE_BUTTON (tracing_t->donttrace_rb)->
-		  active) ? "0" : "1", NULL);
-	  SQLWritePrivateProfileString ("ODBC", "TraceAutoStop",
-	      (GTK_TOGGLE_BUTTON (tracing_t->allthetime_rb)->
-		  active) ? "0" : "1", NULL);
-	  if (STRLEN (gtk_entry_get_text (GTK_ENTRY (tracing_t->
-			  logfile_entry))))
-	    SQLWritePrivateProfileString ("ODBC", "TraceFile",
-		gtk_entry_get_text (GTK_ENTRY (tracing_t->logfile_entry)),
-		NULL);
-	}
+        admin_apply_tracing (tracing_t);
 
       tracing_t->logfile_entry = tracing_t->tracelib_entry =
 	  tracing_t->b_start_stop = NULL;
@@ -830,11 +870,6 @@ delete_event (GtkWidget *widget, GdkEvent *event, void **inparams)
 void
 create_administrator (HWND hwnd)
 {
-  guint b_ok_key, b_cancel_key, b_add_key, b_remove_key, b_test_key,
-      b_configure_key;
-  guint b_donottrace_key, b_allthetime_key, b_onetime_key, b_start_key,
-      b_browse_key, b_select_key;
-  guint br_enable_key, br_disable_key;
   GtkWidget *admin, *dialog_vbox1, *notebook1, *vbox1, *fixed1,
       *scrolledwindow1;
   GtkWidget *clist1, *l_name, *l_description, *l_driver, *l_usdsn, *frame1,
@@ -876,6 +911,7 @@ create_administrator (HWND hwnd)
   TTRACING tracing_t;
   TCONNECTIONPOOLING connectionpool_t;
   void *inparams[6];
+  guint b_key;
 
   if (hwnd == NULL || !GTK_IS_WIDGET (hwnd))
     return;
@@ -1020,10 +1056,10 @@ create_administrator (HWND hwnd)
   gtk_widget_set_usize (vbuttonbox1, 85, 135);
 
   b_add = gtk_button_new_with_label ("");
-  b_add_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
       szDSNButtons[0]);
   gtk_widget_add_accelerator (b_add, "clicked", accel_group,
-      b_add_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_add);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_add", b_add,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1034,10 +1070,10 @@ create_administrator (HWND hwnd)
       'A', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_remove = gtk_button_new_with_label ("");
-  b_remove_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
       szDSNButtons[1]);
   gtk_widget_add_accelerator (b_remove, "clicked", accel_group,
-      b_remove_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_remove);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_remove", b_remove,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1049,11 +1085,11 @@ create_administrator (HWND hwnd)
   gtk_widget_set_sensitive (b_remove, FALSE);
 
   b_configure = gtk_button_new_with_label ("");
-  b_configure_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_configure)->child),
       szDSNButtons[2]);
   gtk_widget_add_accelerator (b_configure, "clicked", accel_group,
-      b_configure_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_configure);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_configure", b_configure,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1065,10 +1101,10 @@ create_administrator (HWND hwnd)
   gtk_widget_set_sensitive (b_configure, FALSE);
 
   b_test = gtk_button_new_with_label ("");
-  b_test_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
       szDSNButtons[3]);
   gtk_widget_add_accelerator (b_test, "clicked", accel_group,
-      b_test_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_test);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_test", b_test,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1199,10 +1235,10 @@ create_administrator (HWND hwnd)
   gtk_widget_set_usize (vbuttonbox2, 85, 135);
 
   b_add = gtk_button_new_with_label ("");
-  b_add_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
       szDSNButtons[0]);
   gtk_widget_add_accelerator (b_add, "clicked", accel_group,
-      b_add_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_add);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_add", b_add,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1213,10 +1249,10 @@ create_administrator (HWND hwnd)
       'A', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_remove = gtk_button_new_with_label ("");
-  b_remove_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
       szDSNButtons[1]);
   gtk_widget_add_accelerator (b_remove, "clicked", accel_group,
-      b_remove_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_remove);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_remove", b_remove,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1228,11 +1264,11 @@ create_administrator (HWND hwnd)
   gtk_widget_set_sensitive (b_remove, FALSE);
 
   b_configure = gtk_button_new_with_label ("");
-  b_configure_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_configure)->child),
       szDSNButtons[2]);
   gtk_widget_add_accelerator (b_configure, "clicked", accel_group,
-      b_configure_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_configure);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_configure", b_configure,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1244,10 +1280,10 @@ create_administrator (HWND hwnd)
   gtk_widget_set_sensitive (b_configure, FALSE);
 
   b_test = gtk_button_new_with_label ("");
-  b_test_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
       szDSNButtons[3]);
   gtk_widget_add_accelerator (b_test, "clicked", accel_group,
-      b_test_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_test);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_test", b_test,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1420,10 +1456,10 @@ create_administrator (HWND hwnd)
      gtk_widget_set_usize (vbuttonbox3, 85, 135);
 
      b_add = gtk_button_new_with_label ("");
-     b_add_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
+     b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
      szDSNButtons[0]);
      gtk_widget_add_accelerator (b_add, "clicked", accel_group,
-     b_add_key, GDK_MOD1_MASK, 0);
+     b_key, GDK_MOD1_MASK, 0);
      gtk_widget_ref (b_add);
      gtk_object_set_data_full (GTK_OBJECT (admin), "b_add", b_add,
      (GtkDestroyNotify) gtk_widget_unref);
@@ -1435,10 +1471,10 @@ create_administrator (HWND hwnd)
      GTK_ACCEL_VISIBLE);
 
      b_remove = gtk_button_new_with_label ("");
-     b_remove_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
+     b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
      szDSNButtons[1]);
      gtk_widget_add_accelerator (b_remove, "clicked", accel_group,
-     b_remove_key, GDK_MOD1_MASK, 0);
+     b_key, GDK_MOD1_MASK, 0);
      gtk_widget_ref (b_remove);
      gtk_object_set_data_full (GTK_OBJECT (admin), "b_remove", b_remove,
      (GtkDestroyNotify) gtk_widget_unref);
@@ -1450,10 +1486,10 @@ create_administrator (HWND hwnd)
      GTK_ACCEL_VISIBLE);
 
      b_configure = gtk_button_new_with_label ("");
-     b_configure_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_configure)->child),
+     b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_configure)->child),
      szDSNButtons[2]);
      gtk_widget_add_accelerator (b_configure, "clicked", accel_group,
-     b_configure_key, GDK_MOD1_MASK, 0);
+     b_key, GDK_MOD1_MASK, 0);
      gtk_widget_ref (b_configure);
      gtk_object_set_data_full (GTK_OBJECT (admin), "b_configure", b_configure,
      (GtkDestroyNotify) gtk_widget_unref);
@@ -1465,10 +1501,10 @@ create_administrator (HWND hwnd)
      GTK_ACCEL_VISIBLE);
 
      b_test = gtk_button_new_with_label ("");
-     b_test_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
+     b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
      szDSNButtons[3]);
      gtk_widget_add_accelerator (b_test, "clicked", accel_group,
-     b_test_key, GDK_MOD1_MASK, 0);
+     b_key, GDK_MOD1_MASK, 0);
      gtk_widget_ref (b_test);
      gtk_object_set_data_full (GTK_OBJECT (admin), "b_test", b_test,
      (GtkDestroyNotify) gtk_widget_unref);
@@ -1615,10 +1651,10 @@ create_administrator (HWND hwnd)
   gtk_button_box_set_child_size (GTK_BUTTON_BOX (hbuttonbox2), 64, -1);
 
   b_add = gtk_button_new_with_label ("");
-  b_add_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
       szDriverButtons[0]);
   gtk_widget_add_accelerator (b_add, "clicked", accel_group,
-      b_add_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_add);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_add", b_add,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1629,10 +1665,10 @@ create_administrator (HWND hwnd)
       'A', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_remove = gtk_button_new_with_label ("");
-  b_remove_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
       szDriverButtons[1]);
   gtk_widget_add_accelerator (b_remove, "clicked", accel_group,
-      b_remove_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_remove);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_remove", b_remove,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1643,11 +1679,11 @@ create_administrator (HWND hwnd)
       'R', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_configure = gtk_button_new_with_label ("");
-  b_configure_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_configure)->child),
       szDriverButtons[2]);
   gtk_widget_add_accelerator (b_configure, "clicked", accel_group,
-      b_configure_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_configure);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_configure", b_configure,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1736,11 +1772,11 @@ create_administrator (HWND hwnd)
   gtk_container_set_border_width (GTK_CONTAINER (vbox5), 6);
 
   br_enable = gtk_radio_button_new_with_label (vbox5_group, "");
-  br_enable_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (br_enable)->child),
       "_Enable");
   gtk_widget_add_accelerator (br_enable, "clicked", accel_group,
-      br_enable_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   vbox5_group = gtk_radio_button_group (GTK_RADIO_BUTTON (br_enable));
   gtk_widget_ref (br_enable);
   gtk_object_set_data_full (GTK_OBJECT (admin), "br_enable", br_enable,
@@ -1749,11 +1785,11 @@ create_administrator (HWND hwnd)
   gtk_box_pack_start (GTK_BOX (vbox5), br_enable, FALSE, FALSE, 0);
 
   br_disable = gtk_radio_button_new_with_label (vbox5_group, "");
-  br_disable_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (br_disable)->child),
       "_Disable");
   gtk_widget_add_accelerator (br_disable, "clicked", accel_group,
-      br_disable_key, GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
+      b_key, GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
   vbox5_group = gtk_radio_button_group (GTK_RADIO_BUTTON (br_disable));
   gtk_widget_ref (br_disable);
   gtk_object_set_data_full (GTK_OBJECT (admin), "br_disable", br_disable,
@@ -1904,11 +1940,11 @@ create_administrator (HWND hwnd)
   gtk_box_pack_start (GTK_BOX (hbox1), vbox2, TRUE, TRUE, 0);
 
   b_donottrace = gtk_radio_button_new_with_label (vbox2_group, "");
-  b_donottrace_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_donottrace)->child),
       "_Don't trace");
   gtk_widget_add_accelerator (b_donottrace, "clicked", accel_group,
-      b_donottrace_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   vbox2_group = gtk_radio_button_group (GTK_RADIO_BUTTON (b_donottrace));
   gtk_widget_ref (b_donottrace);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_donottrace", b_donottrace,
@@ -1920,11 +1956,11 @@ create_administrator (HWND hwnd)
   gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (b_donottrace), TRUE);
 
   b_allthetime = gtk_radio_button_new_with_label (vbox2_group, "");
-  b_allthetime_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_allthetime)->child),
       "All the t_ime");
   gtk_widget_add_accelerator (b_allthetime, "clicked", accel_group,
-      b_allthetime_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   vbox2_group = gtk_radio_button_group (GTK_RADIO_BUTTON (b_allthetime));
   gtk_widget_ref (b_allthetime);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_allthetime", b_allthetime,
@@ -1935,11 +1971,11 @@ create_administrator (HWND hwnd)
       'I', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_onetime = gtk_radio_button_new_with_label (vbox2_group, "");
-  b_onetime_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_onetime)->child),
       "One-_time only");
   gtk_widget_add_accelerator (b_onetime, "clicked", accel_group,
-      b_onetime_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   vbox2_group = gtk_radio_button_group (GTK_RADIO_BUTTON (b_onetime));
   gtk_widget_ref (b_onetime);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_onetime", b_onetime,
@@ -1957,10 +1993,10 @@ create_administrator (HWND hwnd)
   gtk_box_pack_start (GTK_BOX (hbox1), vbox3, FALSE, FALSE, 0);
 
   b_start = gtk_button_new_with_label ("");
-  b_start_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_start)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_start)->child),
       "_Apply tracing settings");
   gtk_widget_add_accelerator (b_start, "clicked", accel_group,
-      b_start_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_start);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_start", b_start,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1997,10 +2033,10 @@ create_administrator (HWND hwnd)
       (GtkAttachOptions) (0), 0, 0);
 
   b_browse = gtk_button_new_with_label ("");
-  b_browse_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_browse)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_browse)->child),
       "_Browse");
   gtk_widget_add_accelerator (b_browse, "clicked", accel_group,
-      b_browse_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_browse);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_browse", b_browse,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -2038,10 +2074,10 @@ create_administrator (HWND hwnd)
       (GtkAttachOptions) (0), 0, 0);
 
   b_select = gtk_button_new_with_label ("");
-  b_select_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_select)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_select)->child),
       "_Select library");
   gtk_widget_add_accelerator (b_select, "clicked", accel_group,
-      b_select_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_select);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_select", b_select,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -2244,9 +2280,9 @@ create_administrator (HWND hwnd)
   gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbuttonbox1), 10);
 
   b_ok = gtk_button_new_with_label ("");
-  b_ok_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_ok)->child), "_Ok");
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_ok)->child), "_Ok");
   gtk_widget_add_accelerator (b_ok, "clicked", accel_group,
-      b_ok_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_ok);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_ok", b_ok,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -2257,10 +2293,10 @@ create_administrator (HWND hwnd)
       'O', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_cancel = gtk_button_new_with_label ("");
-  b_cancel_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_cancel)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_cancel)->child),
       "_Cancel");
   gtk_widget_add_accelerator (b_cancel, "clicked", accel_group,
-      b_cancel_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_cancel);
   gtk_object_set_data_full (GTK_OBJECT (admin), "b_cancel", b_cancel,
       (GtkDestroyNotify) gtk_widget_unref);
