@@ -1,12 +1,29 @@
 /*
  *  dsnchooser.c
  *
- *  $Id: dsnchooser.c,v 1.1 2001/06/07 00:49:40 source Exp $
+ *  $Id: dsnchooser.c,v 1.7 2004/02/19 11:21:20 source Exp $
  *
  *  The iODBC driver manager.
+ *  
+ *  Copyright (C) 1999-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
- *  Copyright (C) 2001 by OpenLink Software <iodbc@openlinksw.com>
+ *  This software is released under the terms of either of the following
+ *  licenses:
  *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -20,17 +37,62 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
+#include <iodbc.h>
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/param.h>
 #include <dirent.h>
 #include <errno.h>
 
-#include "gui.h"
+#include "../gui.h"
 #include "odbc4.xpm"
 
+
+#ifndef MAXPATHLEN
+#define MAXPATHLEN	1024
+#endif
+
+/*
+ *  Linux glibc has the function, but does not have the prototype
+ *  with the default header files.
+ */
+#if defined(HAVE_ASPRINTF) && defined(__GLIBC__)
+extern int asprintf (char **ret, const char *format, ...);
+#endif
+
+
 char *szDSNColumnNames[] = {
   "Name",
   "Description",
@@ -54,25 +116,6 @@ char *szDSNButtons[] = {
   "_Test",
 };
 
-static void
-display_sqlerror (HWND parent, LPCSTR description, LPCSTR dsn, HENV henv,
-    HDBC hdbc, HSTMT hstmt)
-{
-  UCHAR state[32], err[SQL_MAX_MESSAGE_LENGTH];
-  SDWORD native;
-  SWORD len;
-  char *message;
-
-  if (SQLError (henv, hdbc, hstmt, state, &native, err, sizeof (err) - 1,
-	  &len) == SQL_SUCCESS)
-    {
-      message = (char *) malloc (STRLEN (err) + STRLEN ("SQL State : ") + 1);
-      if (message)
-	sprintf (message, "SQL State : %s", state);
-      create_error (parent, dsn, description, message);
-    }
-}
-
 
 void
 addlistofdir_to_optionmenu (GtkWidget *widget, LPCSTR path,
@@ -133,7 +176,11 @@ void
 adddirectories_to_list (HWND hwnd, GtkWidget *widget, LPCSTR path)
 {
   DIR *dir;
-  char *path_buf[MAXPATHLEN * 2];
+#if defined (HAVE_ASPRINTF)
+  char *path_buf;
+#else
+  char path_buf[MAXPATHLEN];
+#endif
   struct dirent *dir_entry;
   struct stat fstat;
   char *data[1];
@@ -148,10 +195,17 @@ adddirectories_to_list (HWND hwnd, GtkWidget *widget, LPCSTR path)
 
       while ((dir_entry = readdir (dir)))
 	{
+#if defined (HAVE_ASPRINTF)
+	  asprintf (&path_buf, "%s/%s", path, dir_entry->d_name);
+#elif defined (HAVE_SNPRINTF)
+	  snprintf (path_buf, sizeof (path_buf), "%s/%s",
+	      path, dir_entry->d_name);
+#else
 	  STRCPY (path_buf, path);
 	  if (path[STRLEN (path) - 1] != '/')
 	    STRCAT (path_buf, "/");
 	  STRCAT (path_buf, dir_entry->d_name);
+#endif
 
 	  if (stat ((LPCSTR) path_buf, &fstat) >= 0
 	      && S_ISDIR (fstat.st_mode))
@@ -160,6 +214,10 @@ adddirectories_to_list (HWND hwnd, GtkWidget *widget, LPCSTR path)
 		data[0] = dir_entry->d_name;
 		gtk_clist_append (GTK_CLIST (widget), data);
 	      }
+
+#if defined (HAVE_ASPRINTF)
+	  free (path_buf);
+#endif
 	}
 
       /* Close the directory entry */
@@ -178,7 +236,11 @@ void
 addfiles_to_list (HWND hwnd, GtkWidget *widget, LPCSTR path)
 {
   DIR *dir;
-  char *path_buf[MAXPATHLEN * 2];
+#if defined (HAVE_ASPRINTF)
+  char *path_buf;
+#else
+  char path_buf[MAXPATHLEN];
+#endif
   struct dirent *dir_entry;
   struct stat fstat;
   char *data[1];
@@ -193,10 +255,17 @@ addfiles_to_list (HWND hwnd, GtkWidget *widget, LPCSTR path)
 
       while ((dir_entry = readdir (dir)))
 	{
+#if defined (HAVE_ASPRINTF)
+	  asprintf (&path_buf, "%s/%s", path, dir_entry->d_name);
+#elif defined (HAVE_SNPRINTF)
+	  snprintf (path_buf, sizeof (path_buf), "%s/%s",
+	      path, dir_entry->d_name);
+#else
 	  STRCPY (path_buf, path);
 	  if (path[STRLEN (path) - 1] != '/')
 	    STRCAT (path_buf, "/");
 	  STRCAT (path_buf, dir_entry->d_name);
+#endif
 
 	  if (stat ((LPCSTR) path_buf, &fstat) >= 0
 	      && !S_ISDIR (fstat.st_mode)
@@ -205,6 +274,10 @@ addfiles_to_list (HWND hwnd, GtkWidget *widget, LPCSTR path)
 	      data[0] = dir_entry->d_name;
 	      gtk_clist_append (GTK_CLIST (widget), data);
 	    }
+
+#if defined (HAVE_ASPRINTF)
+	  free (path_buf);
+#endif
 	}
 
       /* Close the directory entry */
@@ -295,22 +368,29 @@ static void
 dsnchooser_switch_page (GtkNotebook *notebook, GtkNotebookPage *page,
     gint page_num, TDSNCHOOSER *choose_t)
 {
-  if (choose_t)
-    {
-      switch (page_num)
+  switch (page_num)
 	{
 	case 0:
-	  choose_t->type_dsn = 0;
-	  adddsns_to_list (choose_t->udsnlist, FALSE);
+	  if (choose_t)
+	    {
+	      choose_t->type_dsn = 0;
+	      adddsns_to_list (choose_t->udsnlist, FALSE);
+		 }
 	  break;
 
 	case 1:
-	  choose_t->type_dsn = 1;
-	  adddsns_to_list (choose_t->sdsnlist, TRUE);
+	  if (choose_t)
+	    {
+	      choose_t->type_dsn = 1;
+	      adddsns_to_list (choose_t->sdsnlist, TRUE);
+		 }
 	  break;
 
 	case 2:
-	  choose_t->type_dsn = 2;
+	  if (choose_t)
+	    {
+	      choose_t->type_dsn = 2;
+		 }
 	  break;
 	};
 
@@ -329,7 +409,6 @@ dsnchooser_switch_page (GtkNotebook *notebook, GtkNotebookPage *page,
 	  if (choose_t->stest)
 	    gtk_widget_set_sensitive (choose_t->stest, FALSE);
 	}
-    }
 }
 
 
@@ -531,9 +610,8 @@ userdsn_test_clicked (GtkWidget *widget, TDSNCHOOSER *choose_t)
 	  if (SQLAllocHandle (SQL_HANDLE_ENV, NULL, &henv) != SQL_SUCCESS)
 #endif
 	    {
-	      display_sqlerror (choose_t->mainwnd,
-		  "The connection test to the DSN failed.", szDSN,
-		  SQL_NULL_HENV, SQL_NULL_HDBC, SQL_NULL_HSTMT);
+         _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			  henv, SQL_NULL_HDBC, SQL_NULL_HSTMT);
 	      return;
 	    }
 
@@ -545,10 +623,9 @@ userdsn_test_clicked (GtkWidget *widget, TDSNCHOOSER *choose_t)
 	  if (SQLAllocHandle (SQL_HANDLE_DBC, henv, &hdbc) != SQL_SUCCESS)
 #endif
 	    {
+         _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			  henv, hdbc, SQL_NULL_HSTMT);
 	      SQLFreeEnv (henv);
-	      display_sqlerror (choose_t->mainwnd,
-		  "The connection test to the DSN failed.", szDSN, henv,
-		  SQL_NULL_HDBC, SQL_NULL_HSTMT);
 	      return;
 	    }
 
@@ -558,9 +635,8 @@ userdsn_test_clicked (GtkWidget *widget, TDSNCHOOSER *choose_t)
 	  if (SQLDriverConnect (hdbc, choose_t->mainwnd, connstr, SQL_NTS,
 		  outconnstr, sizeof (outconnstr), &buflen,
 		  SQL_DRIVER_PROMPT) != SQL_SUCCESS)
-	    display_sqlerror (choose_t->mainwnd,
-		"The connection test to the DSN failed.", szDSN, henv, hdbc,
-		SQL_NULL_HSTMT);
+        _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			 henv, hdbc, SQL_NULL_HSTMT);
 	  else
 	    {
 	      _iodbcdm_messagebox (choose_t->mainwnd, szDSN,
@@ -787,9 +863,8 @@ systemdsn_test_clicked (GtkWidget * widget, TDSNCHOOSER * choose_t)
 	  if (SQLAllocHandle (SQL_HANDLE_ENV, NULL, &henv) != SQL_SUCCESS)
 #endif
 	    {
-	      display_sqlerror (choose_t->mainwnd,
-		  "The connection test to the DSN failed.", szDSN,
-		  SQL_NULL_HENV, SQL_NULL_HDBC, SQL_NULL_HSTMT);
+         _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			  henv, SQL_NULL_HDBC, SQL_NULL_HSTMT);
 	      return;
 	    }
 
@@ -801,10 +876,9 @@ systemdsn_test_clicked (GtkWidget * widget, TDSNCHOOSER * choose_t)
 	  if (SQLAllocHandle (SQL_HANDLE_DBC, henv, &hdbc) != SQL_SUCCESS)
 #endif
 	    {
+         _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			  henv, hdbc, SQL_NULL_HSTMT);
 	      SQLFreeEnv (henv);
-	      display_sqlerror (choose_t->mainwnd,
-		  "The connection test to the DSN failed.", szDSN, henv,
-		  SQL_NULL_HDBC, SQL_NULL_HSTMT);
 	      return;
 	    }
 
@@ -814,9 +888,8 @@ systemdsn_test_clicked (GtkWidget * widget, TDSNCHOOSER * choose_t)
 	  if (SQLDriverConnect (hdbc, choose_t->mainwnd, connstr, SQL_NTS,
 		  outconnstr, sizeof (outconnstr), &buflen,
 		  SQL_DRIVER_PROMPT) != SQL_SUCCESS)
-	    display_sqlerror (choose_t->mainwnd,
-		"The connection test to the DSN failed.", szDSN, henv, hdbc,
-		SQL_NULL_HSTMT);
+        _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			 henv, hdbc, SQL_NULL_HSTMT);
 	  else
 	    {
 	      _iodbcdm_messagebox (choose_t->mainwnd, szDSN,
@@ -884,27 +957,24 @@ filedsn_add_clicked (GtkWidget *widget, TDSNCHOOSER *choose_t)
 
 	  if (SQLAllocEnv (&henv) != SQL_SUCCESS)
 	    {
-	      display_sqlerror (choose_t->mainwnd,
-		  "The File DSN creation failed.", NULL, SQL_NULL_HENV,
-		  SQL_NULL_HDBC, SQL_NULL_HSTMT);
+         _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			  henv, SQL_NULL_HDBC, SQL_NULL_HSTMT);
 	      return;
 	    }
 
 	  if (SQLAllocConnect (henv, &hdbc) != SQL_SUCCESS)
 	    {
+         _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			  henv, hdbc, SQL_NULL_HSTMT);
 	      SQLFreeEnv (henv);
-	      display_sqlerror (choose_t->mainwnd,
-		  "The File DSN creation failed.", NULL, henv, SQL_NULL_HDBC,
-		  SQL_NULL_HSTMT);
 	      return;
 	    }
 
 	  if (SQLDriverConnect (hdbc, choose_t->mainwnd, connstr, SQL_NTS,
 		  outconnstr, sizeof (outconnstr), &buflen,
 		  SQL_DRIVER_PROMPT) != SQL_SUCCESS)
-	    display_sqlerror (choose_t->mainwnd,
-		"The File DSN creation failed.", NULL, henv, hdbc,
-		SQL_NULL_HSTMT);
+        _iodbcdm_nativeerrorbox (choose_t->mainwnd,
+			 henv, hdbc, SQL_NULL_HSTMT);
 	  else
 	    SQLDisconnect (hdbc);
 
@@ -1178,47 +1248,45 @@ dsnchooser_ok_clicked (GtkWidget *widget, TDSNCHOOSER *choose_t)
   if (choose_t)
     {
       switch (choose_t->type_dsn)
-	{
-	case 0:
-	  if (GTK_CLIST (choose_t->udsnlist)->selection != NULL)
-	    {
-	      gtk_clist_get_text (GTK_CLIST (choose_t->udsnlist),
-		  GPOINTER_TO_INT (GTK_CLIST (choose_t->udsnlist)->selection->
-		      data), 0, &szDSN);
-	      choose_t->dsn = strdup (szDSN);
-	    }
-	  else
-	    choose_t->dsn = NULL;
-	  SQLSetConfigMode (ODBC_USER_DSN);
-	  break;
+	     {
+	       case USER_DSN:
+	         if (GTK_CLIST (choose_t->udsnlist)->selection != NULL)
+	           {
+	             gtk_clist_get_text (GTK_CLIST (choose_t->udsnlist),
+		            GPOINTER_TO_INT (GTK_CLIST (choose_t->udsnlist)->selection->
+		            data), 0, &szDSN);
+	             choose_t->dsn = strdup (szDSN);
+	           }
+	         else
+	           choose_t->dsn = NULL;
+	         break;
 
-	case 1:
-	  if (GTK_CLIST (choose_t->sdsnlist)->selection != NULL)
-	    {
-	      gtk_clist_get_text (GTK_CLIST (choose_t->sdsnlist),
-		  GPOINTER_TO_INT (GTK_CLIST (choose_t->sdsnlist)->selection->
-		      data), 0, &szDSN);
-	      choose_t->dsn = strdup (szDSN);
-	    }
-	  else
-	    choose_t->dsn = NULL;
-	  SQLSetConfigMode (ODBC_SYSTEM_DSN);
-	  break;
+	       case SYSTEM_DSN:
+	         if (GTK_CLIST (choose_t->sdsnlist)->selection != NULL)
+	           {
+	             gtk_clist_get_text (GTK_CLIST (choose_t->sdsnlist),
+		            GPOINTER_TO_INT (GTK_CLIST (choose_t->sdsnlist)->selection->
+		            data), 0, &szDSN);
+	             choose_t->dsn = strdup (szDSN);
+	           }
+	         else
+	           choose_t->dsn = NULL;
+	         break;
 
-	default:
-	  choose_t->dsn = NULL;
-	  break;
-	};
+	       default:
+	         choose_t->dsn = NULL;
+	         break;
+	    };
 
     done:
       choose_t->udsnlist = choose_t->sdsnlist = NULL;
       choose_t->uadd = choose_t->uremove = choose_t->utest =
-	  choose_t->uconfigure = NULL;
+	   choose_t->uconfigure = NULL;
       choose_t->sadd = choose_t->sremove = choose_t->stest =
-	  choose_t->sconfigure = NULL;
+	   choose_t->sconfigure = NULL;
 
       gtk_signal_disconnect_by_func (GTK_OBJECT (choose_t->mainwnd),
-	  GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
+	     GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
       gtk_main_quit ();
       gtk_widget_destroy (choose_t->mainwnd);
     }
@@ -1232,14 +1300,14 @@ dsnchooser_cancel_clicked (GtkWidget *widget, TDSNCHOOSER *choose_t)
     {
       choose_t->udsnlist = choose_t->sdsnlist = NULL;
       choose_t->uadd = choose_t->uremove = choose_t->utest =
-	  choose_t->uconfigure = NULL;
+	   choose_t->uconfigure = NULL;
       choose_t->sadd = choose_t->sremove = choose_t->stest =
-	  choose_t->sconfigure = NULL;
+	   choose_t->sconfigure = NULL;
       choose_t->type_dsn = -1;
       choose_t->dsn = NULL;
 
       gtk_signal_disconnect_by_func (GTK_OBJECT (choose_t->mainwnd),
-	  GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
+	     GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
       gtk_main_quit ();
       gtk_widget_destroy (choose_t->mainwnd);
     }
@@ -1257,8 +1325,6 @@ delete_event (GtkWidget *widget, GdkEvent *event, TDSNCHOOSER *choose_t)
 void
 create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
 {
-  guint b_ok_key, b_cancel_key, b_add_key, b_remove_key, b_test_key,
-      b_configure_key;
   GtkWidget *dsnchooser, *dialog_vbox1, *notebook1, *vbox1, *fixed1,
       *scrolledwindow1;
   GtkWidget *clist1, *l_name, *l_description, *l_driver, *l_usdsn, *frame1,
@@ -1273,6 +1339,13 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   GdkBitmap *mask;
   GtkStyle *style;
   GtkAccelGroup *accel_group;
+  guint b_key;
+
+  if (!GTK_IS_WIDGET (hwnd))
+    {
+      gtk_init(0, NULL);
+      hwnd = gtk_window_new (GTK_WINDOW_TOPLEVEL);
+    }
 
   if (hwnd == NULL || !GTK_IS_WIDGET (hwnd))
     return;
@@ -1282,7 +1355,7 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   dsnchooser = gtk_dialog_new ();
   gtk_object_set_data (GTK_OBJECT (dsnchooser), "dsnchooser", dsnchooser);
   gtk_widget_set_usize (dsnchooser, 565, 415);
-  gtk_window_set_title (GTK_WINDOW (dsnchooser), "Select a Data Source");
+  gtk_window_set_title (GTK_WINDOW (dsnchooser), "Select Data Source");
   gtk_window_set_position (GTK_WINDOW (dsnchooser), GTK_WIN_POS_CENTER);
   gtk_window_set_modal (GTK_WINDOW (dsnchooser), TRUE);
   gtk_window_set_policy (GTK_WINDOW (dsnchooser), FALSE, FALSE, FALSE);
@@ -1305,6 +1378,7 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_widget_show (vbox1);
   gtk_container_add (GTK_CONTAINER (notebook1), vbox1);
 
+  /* User DSN panel */
   fixed1 = gtk_fixed_new ();
   gtk_widget_ref (fixed1);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "fixed1", fixed1,
@@ -1415,10 +1489,10 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_widget_set_usize (vbuttonbox1, 85, 135);
 
   b_add = gtk_button_new_with_label ("");
-  b_add_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
       szDSNButtons[0]);
   gtk_widget_add_accelerator (b_add, "clicked", accel_group,
-      b_add_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_add);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_add", b_add,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1429,10 +1503,10 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
       'A', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_remove = gtk_button_new_with_label ("");
-  b_remove_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
       szDSNButtons[1]);
   gtk_widget_add_accelerator (b_remove, "clicked", accel_group,
-      b_remove_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_remove);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_remove", b_remove,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1444,11 +1518,11 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_widget_set_sensitive (b_remove, FALSE);
 
   b_configure = gtk_button_new_with_label ("");
-  b_configure_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_configure)->child),
       szDSNButtons[2]);
   gtk_widget_add_accelerator (b_configure, "clicked", accel_group,
-      b_configure_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_configure);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_configure",
       b_configure, (GtkDestroyNotify) gtk_widget_unref);
@@ -1460,10 +1534,10 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_widget_set_sensitive (b_configure, FALSE);
 
   b_test = gtk_button_new_with_label ("");
-  b_test_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
       szDSNButtons[3]);
   gtk_widget_add_accelerator (b_test, "clicked", accel_group,
-      b_test_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_test);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_test", b_test,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1487,6 +1561,7 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_notebook_set_tab_label (GTK_NOTEBOOK (notebook1),
       gtk_notebook_get_nth_page (GTK_NOTEBOOK (notebook1), 0), udsn);
 
+  /* System DSN panel */
   fixed2 = gtk_fixed_new ();
   gtk_widget_ref (fixed2);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "fixed2", fixed2,
@@ -1593,10 +1668,10 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_widget_set_usize (vbuttonbox2, 85, 135);
 
   b_add = gtk_button_new_with_label ("");
-  b_add_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_add)->child),
       szDSNButtons[0]);
   gtk_widget_add_accelerator (b_add, "clicked", accel_group,
-      b_add_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_add);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_add", b_add,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1607,10 +1682,10 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
       'A', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_remove = gtk_button_new_with_label ("");
-  b_remove_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_remove)->child),
       szDSNButtons[1]);
   gtk_widget_add_accelerator (b_remove, "clicked", accel_group,
-      b_remove_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_remove);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_remove", b_remove,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1622,11 +1697,11 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_widget_set_sensitive (b_remove, FALSE);
 
   b_configure = gtk_button_new_with_label ("");
-  b_configure_key =
+  b_key =
       gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_configure)->child),
       szDSNButtons[2]);
   gtk_widget_add_accelerator (b_configure, "clicked", accel_group,
-      b_configure_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_configure);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_configure",
       b_configure, (GtkDestroyNotify) gtk_widget_unref);
@@ -1638,10 +1713,10 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_widget_set_sensitive (b_configure, FALSE);
 
   b_test = gtk_button_new_with_label ("");
-  b_test_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_test)->child),
       szDSNButtons[3]);
   gtk_widget_add_accelerator (b_test, "clicked", accel_group,
-      b_test_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_test);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_test", b_test,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1682,9 +1757,9 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
   gtk_button_box_set_spacing (GTK_BUTTON_BOX (hbuttonbox1), 10);
 
   b_ok = gtk_button_new_with_label ("");
-  b_ok_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_ok)->child), "_Ok");
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_ok)->child), "_Ok");
   gtk_widget_add_accelerator (b_ok, "clicked", accel_group,
-      b_ok_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_ok);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_ok", b_ok,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1695,10 +1770,10 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
       'O', GDK_MOD1_MASK, GTK_ACCEL_VISIBLE);
 
   b_cancel = gtk_button_new_with_label ("");
-  b_cancel_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_cancel)->child),
+  b_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (b_cancel)->child),
       "_Cancel");
   gtk_widget_add_accelerator (b_cancel, "clicked", accel_group,
-      b_cancel_key, GDK_MOD1_MASK, 0);
+      b_key, GDK_MOD1_MASK, 0);
   gtk_widget_ref (b_cancel);
   gtk_object_set_data_full (GTK_OBJECT (dsnchooser), "b_cancel", b_cancel,
       (GtkDestroyNotify) gtk_widget_unref);
@@ -1763,7 +1838,7 @@ create_dsnchooser (HWND hwnd, TDSNCHOOSER * choose_t)
 
   choose_t->udsnlist = clist1;
   choose_t->sdsnlist = clist2;
-  choose_t->type_dsn = 0;
+  choose_t->type_dsn = USER_DSN;
   choose_t->mainwnd = dsnchooser;
 
   gtk_widget_show_all (dsnchooser);
