@@ -1,12 +1,29 @@
 /*
  *  gui.h
  *
- *  $Id: gui.h,v 1.1 2001/06/07 00:49:40 source Exp $
+ *  $Id: gui.h,v 1.3 2003/08/22 14:28:42 source Exp $
  *
  *  The iODBC driver manager.
+ *  
+ *  Copyright (C) 1999-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
- *  Copyright (C) 2001 by OpenLink Software <iodbc@openlinksw.com>
+ *  This software is released under the terms of either of the following
+ *  licenses:
  *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -20,6 +37,35 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <iodbc.h>
@@ -114,7 +160,8 @@ void filedsn_filelist_select(GtkWidget* widget, gint row, gint column, GdkEvent
 void filedsn_filelist_unselect(GtkWidget* widget, gint row, gint column, GdkEvent *event, TDSNCHOOSER *choose_t);
 void filedsn_dirlist_select(GtkWidget* widget, gint row, gint column, GdkEvent *event, TDSNCHOOSER *choose_t);
 void filedsn_lookin_clicked(GtkWidget* widget, void **array);
-void adddrivers_to_list(GtkWidget* widget, BOOL isTrs);
+void adddrivers_to_list(GtkWidget* widget, GtkWidget* dlg);
+void addtranslators_to_list(GtkWidget* widget, GtkWidget* dlg);
 void adddirectories_to_list(HWND hwnd, GtkWidget* widget, LPCSTR path);
 void addfiles_to_list(HWND hwnd, GtkWidget* widget, LPCSTR path);
 void addlistofdir_to_optionmenu(GtkWidget* widget, LPCSTR path, TDSNCHOOSER *choose_t);
