@@ -1,12 +1,29 @@
 /*
  *  translator.c
  *
- *  $Id: translatorchooser.c,v 1.1 2001/06/07 00:49:40 source Exp $
+ *  $Id: translatorchooser.c,v 1.3 2003/08/22 14:28:42 source Exp $
  *
  *  The iODBC driver manager.
+ *  
+ *  Copyright (C) 1999-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
- *  Copyright (C) 2001 by OpenLink Software <iodbc@openlinksw.com>
+ *  This software is released under the terms of either of the following
+ *  licenses:
  *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -20,6 +37,35 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <sys/stat.h>
@@ -100,6 +146,118 @@ delete_event (GtkWidget *widget,
   return FALSE;
 }
 
+void
+addtranslators_to_list (GtkWidget *widget, GtkWidget *dlg)
+{
+  char *curr, *buffer = (char *) malloc (sizeof (char) * 65536), *szDriver;
+  char driver[1024], _date[1024], _size[1024];
+  char *data[4];
+  int len, row = 0, i;
+  BOOL careabout;
+  UWORD confMode = ODBC_USER_DSN;
+  struct stat _stat;
+
+  if (!buffer || !GTK_IS_CLIST (widget))
+    return;
+  gtk_clist_clear (GTK_CLIST (widget));
+
+  /* Get the current config mode */
+  while (confMode != ODBC_SYSTEM_DSN + 1)
+    {
+      /* Get the list of drivers in the user context */
+      SQLSetConfigMode (confMode);
+#ifdef WIN32
+      len =
+	  SQLGetPrivateProfileString ("ODBC 32 bit Translators",
+	   	NULL, "", buffer, 65535, "odbcinst.ini");
+#else
+      len =
+	  SQLGetPrivateProfileString ("ODBC Translators",
+	   	NULL, "", buffer, 65535, "odbcinst.ini");
+#endif
+      if (len)
+	goto process;
+
+      goto end;
+
+    process:
+      for (curr = buffer; *curr; curr += (STRLEN (curr) + 1))
+	{
+	  /* Shadowing system odbcinst.ini */
+	  for (i = 0, careabout = TRUE; i < GTK_CLIST (widget)->rows; i++)
+	    {
+	      gtk_clist_get_text (GTK_CLIST (widget), i, 0, &szDriver);
+	      if (!strcmp (szDriver, curr))
+		{
+		  careabout = FALSE;
+		  break;
+		}
+	    }
+
+	  if (!careabout)
+	    continue;
+
+	  SQLSetConfigMode (confMode);
+#ifdef WIN32
+	  SQLGetPrivateProfileString ("ODBC 32 bit Translators",
+	      curr, "", driver, sizeof (driver),
+	      "odbcinst.ini");
+#else
+	  SQLGetPrivateProfileString ("ODBC Translators",
+	      curr, "", driver, sizeof (driver),
+	      "odbcinst.ini");
+#endif
+
+	  /* Check if the driver is installed */
+	  if (strcasecmp (driver, "Installed"))
+	    goto end;
+
+	  /* Get the driver library name */
+	  SQLSetConfigMode (confMode);
+	  if (!SQLGetPrivateProfileString (curr,
+		  "Translator", "", driver,
+		  sizeof (driver), "odbcinst.ini"))
+	    {
+	      SQLSetConfigMode (confMode);
+	      SQLGetPrivateProfileString ("Default",
+		     "Translator", "", driver,
+		  sizeof (driver), "odbcinst.ini");
+	    }
+
+	  if (STRLEN (curr) && STRLEN (driver))
+	    {
+	      data[0] = curr;
+	      data[1] = driver;
+
+	      /* Get some information about the driver */
+	      if (!stat (driver, &_stat))
+		{
+		  sprintf (_size, "%d Kb", _stat.st_size / 1024);
+		  sprintf (_date, "%s", ctime (&_stat.st_mtime));
+		  data[2] = _date;
+		  data[3] = _size;
+		  gtk_clist_append (GTK_CLIST (widget), data);
+		}
+	    }
+	}
+
+    end:
+      if (confMode == ODBC_USER_DSN)
+	confMode = ODBC_SYSTEM_DSN;
+      else
+	confMode = ODBC_SYSTEM_DSN + 1;
+    }
+
+  if (GTK_CLIST (widget)->rows > 0)
+    {
+      gtk_clist_columns_autosize (GTK_CLIST (widget));
+      gtk_clist_sort (GTK_CLIST (widget));
+    }
+
+  /* Make the clean up */
+  free (buffer);
+}
+
 
 void
 create_translatorchooser (HWND hwnd, TTRANSLATORCHOOSER *choose_t)
@@ -275,7 +433,7 @@ create_translatorchooser (HWND hwnd, TTRANSLATORCHOOSER *choose_t)
 
   gtk_window_add_accel_group (GTK_WINDOW (translatorchooser), accel_group);
 
-  adddrivers_to_list (clist1, TRUE);
+  addtranslators_to_list (clist1, translatorchooser);
 
   choose_t->translatorlist = clist1;
   choose_t->translator = NULL;
