@@ -1,14 +1,31 @@
 /*
  *  misc.c
  *
- *  $Id: misc.c,v 1.1 2001/06/07 00:48:50 source Exp $
+ *  $Id: misc.c,v 1.7 2004/02/18 15:57:50 source Exp $
  *
  *  Miscellaneous functions
  *
  *  The iODBC driver manager.
+ *  
+ *  Copyright (C) 1999-2002 by OpenLink Software <iodbc@openlinksw.com>
+ *  All Rights Reserved.
  *
- *  Copyright (C) 2001 by OpenLink Software <iodbc@openlinksw.com>
+ *  This software is released under the terms of either of the following
+ *  licenses:
  *
+ *      - GNU Library General Public License (see LICENSE.LGPL) 
+ *      - The BSD License (see LICENSE.BSD).
+ *
+ *  While not mandated by the BSD license, any patches you make to the
+ *  iODBC source code may be contributed back into the iODBC project
+ *  at your discretion. Contributions will benefit the Open Source and
+ *  Data Access community as a whole. Submissions may be made at:
+ *
+ *      http://www.iodbc.org
+ *
+ *
+ *  GNU Library Generic Public License Version 2
+ *  ============================================
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Library General Public
  *  License as published by the Free Software Foundation; either
@@ -22,6 +39,35 @@
  *  You should have received a copy of the GNU Library General Public
  *  License along with this library; if not, write to the Free
  *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *
+ *
+ *  The BSD License
+ *  ===============
+ *  Redistribution and use in source and binary forms, with or without
+ *  modification, are permitted provided that the following conditions
+ *  are met:
+ *
+ *  1. Redistributions of source code must retain the above copyright
+ *     notice, this list of conditions and the following disclaimer.
+ *  2. Redistributions in binary form must reproduce the above copyright
+ *     notice, this list of conditions and the following disclaimer in
+ *     the documentation and/or other materials provided with the
+ *     distribution.
+ *  3. Neither the name of OpenLink Software Inc. nor the names of its
+ *     contributors may be used to endorse or promote products derived
+ *     from this software without specific prior written permission.
+ *
+ *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
+ *  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL OPENLINK OR
+ *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
+ *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
+ *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
+ *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
+ *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
+ *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+ *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <iodbc.h>
@@ -31,6 +77,8 @@
 #include <stdlib.h>
 #include <string.h>
 #include <unistd.h>
+#include <sys/types.h>
+#include <sys/stat.h>
 
 #include "inifile.h"
 #include "misc.h"
@@ -46,7 +94,7 @@ static int
 upper_strneq (char *s1, char *s2, int n)
 {
   int i;
-  char c1, c2;
+  char c1 = '\0', c2 = '\0';
 
   for (i = 1; i < n; i++)
     {
@@ -155,11 +203,11 @@ readtoken (
  * For MacX:     1. Check for $ODBCINI variable, if exists return $ODBCINI.
  *               2. Check for $HOME/.odbc.ini or ~/.odbc.ini file, if exists
  *                  return it.
- *               3. Check for $HOME/Library/Preferences/ODBC.preference or
+ *               3. Check for $HOME/Library/ODBC/odbc.ini or
  *                  ~/.odbc.ini file, if exists return it.
  *               4. Check for SYS_ODBC_INI build variable, if exists return
  *                  it. (ie : /etc/odbc.ini).
- *               5. Check for /System/Library/Preferences/ODBC.preference
+ *               5. Check for /Library/ODBC/odbc.ini
  *                  file, if exists return it.
  *               6. No odbc.ini presence, return NULL.
  */
@@ -167,9 +215,10 @@ char *
 _iodbcadm_getinifile (char *buf, int size, int bIsInst, int doCreate)
 {
 #ifdef _MAC
-  OSErr result;
+  HParamBlockRec hp;
   long fldrDid;
   short fldrRef;
+  OSErr result;
 #endif /* endif _MAC */
   int i, j;
   char *ptr;
@@ -215,6 +264,22 @@ _iodbcadm_getinifile (char *buf, int size, int bIsInst, int doCreate)
   STRCAT (buf, bIsInst ? ":ODBC Installer Preferences" : ":ODBC Preferences");
 #    endif /* endif __POWERPC__ */
 
+  if (doCreate)
+    {
+      hp.fileParam.ioCompletion = NULL;
+      hp.fileParam.ioVRefNum = fldrRef;
+      hp.fileParam.ioDirID = fldrDid;
+#    ifdef __POWERPC__
+      hp.fileParam.ioNamePtr =
+	  bIsInst ? "\pODBC Installer Preferences PPC" :
+	  "\pODBC Preferences PPC";
+#    else
+      hp.fileParam.ioNamePtr =
+	  bIsInst ? "\pODBC Installer Preferences" : "\pODBC Preferences";
+#    endif
+      PBHCreate (&hp, FALSE);
+    }
+
   free (ptr);
 
   return buf;
@@ -243,8 +308,18 @@ _iodbcadm_getinifile (char *buf, int size, int bIsInst, int doCreate)
 	{
 	  STRNCPY (buf, ptr, size);
 
-	  if (doCreate || access (buf, R_OK) == 0)
-	    return buf;
+          if (access (buf, R_OK) == 0)
+            return buf;
+          else if (doCreate)
+            {
+              int f = open ((char*)buf, O_CREAT,
+                S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+              if(f != -1)
+                {
+                  close(f);
+                  return buf;
+                }
+            }
 	}
 
 #  ifdef VMS
@@ -276,16 +351,26 @@ _iodbcadm_getinifile (char *buf, int size, int bIsInst, int doCreate)
 	  if (doCreate || access (buf, R_OK) == 0)
 	    return buf;
 
-#   ifdef _MACX
+#   ifdef __APPLE__
 	  /*
-	   * Try to check the ~/Library/Preferences/ODBC.preference
+	   * Try to check the ~/Library/ODBC/odbc.ini
 	   */
 	  snprintf (buf, size,
 	      bIsInst ? "%s" ODBCINST_INI_APP : "%s" ODBC_INI_APP, ptr);
 
-	  if (doCreate || access (buf, R_OK) == 0)
-	    return buf;
-#   endif /* endif _MACX */
+          if (access (buf, R_OK) == 0)
+            return buf;
+          else if (doCreate)
+            {
+              int f = open ((char*)buf, O_CREAT,
+                S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+              if(f != -1)
+                {
+                  close(f);
+                  return buf;
+                }
+            }
+#   endif /* endif __APPLE__ */
 	}
 
 #  endif /* endif VMS */
@@ -303,25 +388,43 @@ _iodbcadm_getinifile (char *buf, int size, int bIsInst, int doCreate)
 	{
 	  STRNCPY (buf, ptr, size);
 
-	  if (doCreate || access (buf, R_OK) == 0)
-	    return buf;
+          if (access (buf, R_OK) == 0)
+            return buf;
+          else if (doCreate)
+            {
+              int f = open ((char*)buf, O_CREAT,
+                S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+              if(f != -1)
+                {
+                  close(f);
+                  return buf;
+                }
+            }
 	}
 
-      STRNCPY (buf, bIsInst ? SYS_ODBCINST_INI : SYS_ODBC_INI, size);
-
-      if (doCreate || access (buf, R_OK) == 0)
-	return buf;
-
-#   ifdef _MACX
+#   ifdef __APPLE__
       /*
-       * Try to check the /System/Library/Preferences/ODBC.preference
+       * Try to check the /Library/ODBC/odbc.ini
        */
-      snprintf (buf, size, "/System%s",
+      snprintf (buf, size, "%s",
 	  bIsInst ? ODBCINST_INI_APP : ODBC_INI_APP);
 
       if (access (buf, R_OK) == 0)
-	return buf;
-#   endif /* endif _MACX */
+        return buf;
+      else if (doCreate)
+        {
+          int f = open ((char*)buf, O_CREAT,
+            S_IREAD | S_IWRITE | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH);
+          if(f != -1)
+            {
+              close(f);
+              return buf;
+            }
+        }
+#   endif /* endif __APPLE__ */
+
+      STRNCPY (buf, bIsInst ? SYS_ODBCINST_INI : SYS_ODBC_INI, size);
+      return buf;
     }
 
   /*
@@ -335,7 +438,7 @@ _iodbcadm_getinifile (char *buf, int size, int bIsInst, int doCreate)
 const char *
 _iodbcdm_check_for_string(const char *szList, const char *szString, int bContains)
 {
-  const char *currP = szList, *ret;
+  const char *currP = szList;
 
   while (*currP)
     {
